//
// GameX (Release 2 - Vectors / Matricies / CameraX)
// CameraX Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
//
// 3D CAMERA IMPLEMENTATION 
// 
// The camera transformation of an arbitrary point is:
//
//		Q' = Q * T * R * P
//
// where Q	= 3D point
//		 Q' = Screen point
//		 T = Camera Translation (moves camera to origin)
//		 R = Camera Rotation (rotates camera to point 'up' along Z axis)
//		 P = Projection (projects points onto XY plane)
// 
// T is a unit-coordinate system translated to origin from camera:
//		[1	0  0  0]
//		[0	1  0  0]
//		[0	0  1  0]
//		[-cx -cy -cz 0] 	where c is camera location
// R is a z_basis matrix: 
//
// P is a projection matrix:
//
	
#include "gamex-camera.h"

CameraX::CameraX ()
{	
	Vec3f from (-150, -150, 40);	// Default camera
	Vec3f to (0, 0, 0);
	Vec3f up (0, 0, 1);
	
	z_dir_vector_invalid = true;
	z_proj_matrix_invalid = true;
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;

	z_proj_type = CAMERA_PERSPECT;
	z_near_dist = 1.0f;
	z_far_dist = 1000.0f;
	z_plane_sep = 1.0f / (z_far_dist - z_near_dist);
	z_height_dist = 1.0f;
	z_near_clip = 0.0f;
	z_roll = 0.0f;
	z_offset_x = 0.0f;
	z_offset_y = 0.0f;
	SetFromPosition (from);
	SetToPosition (to);
	SetUpVector (up);
	SetWindow (0, 0, 800, 600);

	z_MakeTransformMatrix();
}

void CameraX::z_MakeDirVector () const
{
	z_dir_vector = z_to_pos;
	z_dir_vector -= z_from_pos;
	z_dir_vector.Normalize();
	z_dir_vector_invalid = false;
}


void CameraX::z_MakeViewMatrix () const
{
	if(z_dir_vector_invalid) z_MakeDirVector();

	Vec3f up = z_up_vector;
	Vec3f dir = z_dir_vector;
	Vec3f temp;

	z_view_matrix = Mat44f::IdentityMatrix;
	z_view_matrix(3, 0) = -z_from_pos.x;
	z_view_matrix(3, 1) = -z_from_pos.y;
	z_view_matrix(3, 2) = -z_from_pos.z;

	if(z_roll != 0.0f)
	{
		z_rollM = z_rollM.RotateZMatrix(sinf(-z_roll),cosf(z_roll));
		up *= z_rollM;
	}

	temp = dir;
	temp *= up.Dot(dir);
	up -= temp;
	up.Normalize();

	z_side_vector = dir;
	z_side_vector.CrossWith (up);
	z_basis.Basis (z_side_vector, up, dir);
	
	z_view_matrix *= z_basis;

	z_view_matrix_invalid = false;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::z_MakeProjMatrix () const
{
	switch (z_proj_type)
	{
		case CAMERA_PERSPECT:
		{
			z_proj_matrix = 0.0f;
			z_proj_matrix(0,0) = z_near_dist/z_height_dist;
			z_proj_matrix(1,1) = z_near_dist/z_height_dist;
			z_proj_matrix(2,2) = z_far_dist/(z_far_dist-z_near_dist);	
			z_proj_matrix(3,2) = (-z_near_dist*z_far_dist)/(z_far_dist-z_near_dist);
			z_proj_matrix(2,3) = 1.0f;
		}	break;
	}

	z_proj_matrix_invalid = false;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::z_MakeTransformMatrix () const
{
	if(z_view_matrix_invalid) z_MakeViewMatrix();
	if(z_proj_matrix_invalid) z_MakeProjMatrix();

	z_transform_matrix = z_view_matrix;
	z_transform_matrix *= z_proj_matrix;

//	z_transform_matrix_invalid = false;
}

void CameraX::SetFromPosition (const Vec3f &from)
{
	z_from_pos = from;	
	z_dir_vector_invalid = true;
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetFromAngles (const Vec3f &angs)
{
	float ang = angs.x*DEGtoRAD;
	float tilt = angs.y*DEGtoRAD;
	float dolly = angs.z;
	float cosTilt = cosf(tilt);
	z_from_pos.x = z_to_pos.x - (sinf(ang)*cosTilt*dolly);
	z_from_pos.y = z_to_pos.y - (cosf(ang)*cosTilt*dolly);
	z_from_pos.z = z_to_pos.z + (sinf(tilt)*dolly);
	z_dir_vector_invalid = true;
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetToPosition (const Vec3f &to)
{
	z_to_pos = to;	
	z_dir_vector_invalid = true;
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetUpVector (const Vec3f &up)
{
	z_up_vector = up; 
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetRollAngle (float angle)
{
	z_roll = angle * -DEGtoRAD;
	z_view_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetNearPlane (float n)
{
	z_near_dist = n;
	z_plane_sep = 1.0f / (z_far_dist - z_near_dist);
	z_proj_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetFarPlane (float f)
{
	z_far_dist = f;
	z_plane_sep = 1.0f / (z_far_dist - z_near_dist);
	z_proj_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetViewHeight (float h)
{
	z_height_dist = h;
	z_proj_matrix_invalid = true;
//	z_transform_matrix_invalid = true;
	z_MakeTransformMatrix();
}

void CameraX::SetProjection (int proj_const)
{
	z_proj_type = proj_const;
}

void CameraX::SetWindow (int x1, int y1, int x2, int y2)
{
	z_cam_x1 = (float)x1; 				z_cam_y1 = (float)y1;
	z_cam_x2 = (float)x2; 				z_cam_y2 = (float)y2;
	z_cam_sx = z_cam_x2 - z_cam_x1;			z_cam_sy = z_cam_y2 - z_cam_y1;
	z_cam_cx = z_cam_x1 + z_cam_sx/2.0f;		z_cam_cy = z_cam_y1 + z_cam_sy/2.0f;
}

void CameraX::GetWindow (int&x1, int&y1, int&x2, int&y2)
{
	x1 = Round(z_cam_x1);
	y1 = Round(z_cam_y1);
	x2 = Round(z_cam_x2);
	y2 = Round(z_cam_y2);
}


//*** Projection Pipelines:
//
//	z_view_matrix =			Rotation and Translation z_basis for Camera
//	z_proj_matrix =			Perspective (or other) Projection matrix
//	z_transform_matrix =		z_view_matrix * z_proj_matrix
//
// Standard Projection Pipelines:
//
//		z_view_matrix 		  z_proj_matrix		  4D to 3D vector
//	p --ViewTransform--> q --ProjTransform--> q --ScreenTransform --> p
//		(returns projected point p in range -1 < x, y, z < 1)
//
// Window Projection Pipeline:
//
//		z_view_matrix 		  z_proj_matrix		  4D to 3D vector		  Remap to Window
//	p --ViewTransform--> q --ProjTransform--> q --ScreenTransform --> q --WindowTransform--> (x,y)
//		(returns points in range -win < x, y < win, 0 < z < far)
//
// Combined View/Project Pipeline:
//
//		z_transform_matrix	   Remap to Window
//	p --CameraTransform--> q --WindowTransform--> (x,y)

void CameraX::ViewTransform (Vec3f &p, Vec4f &q)
{
	q = p;	
	q.w = 1;
	q *= z_view_matrix;
}

void CameraX::ProjectTransform (Vec4f &q)
{
	if(z_proj_matrix_invalid) z_MakeProjMatrix();
	q *= z_proj_matrix;
}

void CameraX::CameraTransform (Vec3f &p, Vec4f &q)
{
	if(z_transform_matrix_invalid) z_MakeTransformMatrix();
	q = p;
	q *= z_transform_matrix;
}

void CameraX::ScreenTransform (Vec4f &q, Vec3f &p)
{
	p = q;
}

void CameraX::WindowTransform (Vec4f &q, float &x, float &y, float &z)
{
	float sx, sy;

	sx = z_cam_x2 - z_cam_x1;
	sy = z_cam_y2 - z_cam_y1;
	x = sx/2.0f + z_cam_x1 + (sx*q.x/q.w);
	y = sy/2.0f + z_cam_y1 - (sy*q.y/q.w);
	z = q.z / q.w;
}

// Inverse Pipeline:
//
//
//	p -->InverseTransform--> q
//		(given points in range -1 < x, y < 1, z = fixed at near)
//

//		z_view_matrix 		  z_proj_matrix		  4D to 3D vector
//	p --ViewTransform--> q --ProjTransform--> q --ScreenTransform --> p
//		(returns projected point p in range -1 < x, y, z < 1)


void CameraX::InverseTransform (Vec3f &p, Vec4f &q)
{
	Vec3f up = z_up_vector;
	Vec3f temp = z_dir_vector;

	temp *= z_up_vector.Dot(z_dir_vector);
	up -= temp;
	up.Normalize(); 
	z_side_vector = z_dir_vector;
	z_side_vector.CrossWith (up);
	
	z_inv_view.Basis (z_side_vector, up, z_dir_vector);
	z_inv_view.Transpose ();

	z_inv_proj = 0.0f;
	z_inv_proj(0,0) = z_height_dist/z_near_dist;
	z_inv_proj(1,1) = z_height_dist/z_near_dist;
	z_inv_proj(2,2) = (z_far_dist-z_near_dist)/z_far_dist;
	z_inv_proj(3,2) = (z_far_dist-z_near_dist)/(-z_near_dist*z_far_dist);
	z_inv_proj(2,3) = 1.0f;

	p.z = z_near_dist;
	q = p; 
//	q *= (float) (-z_near_dist*z_far_dist / (p.Z()*(z_far_dist-z_near_dist)) + z_near_dist);
	q *= z_inv_proj;
	q *= z_inv_view;
}

// Accelerated Pipeline:
//
//	  p -> (4DF) q -> z_transform_matrix -> window_transform -> (x, y, z)
//		(returns points in range:
//				 camx1 < x < camx2
//				 camy1 < y < camy2
//					 0 < z < far)
//
void CameraX::Project (const Vec3f &p, float &x, float &y, float &z)
{
	const float * m = z_transform_matrix.GetData();

	float qw = p.x * m[3] + p.y * m[7] + p.z * m[11] + m[15];
	if(qw < 0.000001f)
		qw = 0.000001f; // avoid divide-by-zero and divide-by-negative
	const float csx = z_cam_sx/qw;

	x = z_cam_cx + (csx * (p.x * m[0] + p.y * m[4] + p.z * m[8] + m[12])) + z_offset_x;
	y = z_cam_cy - (csx * (p.x * m[1] + p.y * m[5] + p.z * m[9] + m[13])) + z_offset_y;
	z =                    p.x * m[2] + p.y * m[6] + p.z * m[10] + m[14];
	if(z > z_far_dist)
		z = -z;
}

void CameraX::ProjectFast (register const Vec3f &p, float &x, float &y, float &z)
{
	const float * m = z_transform_matrix.GetData();
	const float csx = z_cam_sx / (p.x * m[3] + p.y * m[7] + p.z * m[11] + m[15]);
	x = z_cam_cx + (csx * (p.x * m[0] + p.y * m[4] + p.z * m[8]  + m[12]));
	y = z_cam_cy - (csx * (p.x * m[1] + p.y * m[5] + p.z * m[9]  + m[13]));
	z =                    p.x * m[2] + p.y * m[6] + p.z * m[10] + m[14];
}


/* // old version - maybe clearer, but slower
void CameraX::Project (const Vec3f &p, float &x, float &y, float &z)
{
	if(z_transform_matrix_invalid) z_MakeTransformMatrix();
	float *m = z_transform_matrix.GetData ();
	float qx, qy, qz, qw;
		
	// Explicit vector4DF-Mat44f multiplication (p * z_transform_matrix) 
	qx	= p.x * (*m++); qy	= p.x * (*m++); qz	= p.x * (*m++); qw	= p.x * (*m++);
	qx += p.y * (*m++); qy += p.y * (*m++); qz += p.y * (*m++); qw += p.y * (*m++);
	qx += p.z * (*m++); qy += p.z * (*m++); qz += p.z * (*m++); qw += p.z * (*m++);
	qx +=		(*m++); qy +=		(*m++); qz +=		(*m++); qw +=		(*m  );

	if(qw < 0.000001f) qw = 0.000001f; // avoid divide-by-zero and divide-by-negative

	float csx = z_cam_sx/qw;
	x = z_cam_cx + (csx * qx) + z_offset_x;
	y = z_cam_cy - (csx * qy) + z_offset_y;
	z = qz;
	if(z>z_far_dist) z = -z;
}
*/