//
// GameX - BMP/JPG/GIF/PNG-Handling Code
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
//
//**************************************
//
// BMP Format - File Format support for ImageExtender class
//
// NOTE: CURRENT CAPABILITIES:
//		- Load supports:
//					 8 bit indexed color    (no alpha)
//					16 bit X1-R5-G5-B5 (without alpha)
//					24 bit    R8-G8-B8 (without alpha)
//					32 bit A8-R8-G8-B8 (with alpha)
//		- Save supports:
//					16 bit X1-R5-G5-B5 (without alpha)
//					32 bit A8-R8-G8-B8 (with alpha)
//
//	Other Formats: Can load JPG and GIF files (without alpha) or PNG files (with or without alpha)
//
//**************************************

#include "gamex-image.h"
#include "gamex-win-dx.h"

#include <olectl.h> // required for JPG/GIF reading
#include <ocidl.h>

// (also need to link with ole32.lib and oleaut32.lib)
#pragma comment(lib,"ole32.lib") // auto-link ole32.lib
#pragma comment(lib,"oleaut32.lib ") // auto-link oleaut32.lib 


// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif


char last_filename [MAX_PATH];



// Function: LoadBMP
//
//		filename			Name of BMP file to load
//		img					ImageX object to fill with image data from bitmap 

bool ImageExt::LoadBMP (const char * filename, ImageX *img)
{
///	if(strcmp(last_filename, filename) == 0)  // sometimes Windows inserts GUI element graphics where the BMP should be
///		Sleep(120); // when loading the same file twice in a row without waiting a bit in-between

	m_pImg = img;
	strcpy(m_filename, filename);

//	debug.FixPath(m_filename);

	strcpy(last_filename, m_filename);

	// actually load the bitmap from file:
	HBITMAP hbmp = (HBITMAP) LoadImage(NULL,m_filename,IMAGE_BITMAP,0,0,LR_LOADFROMFILE|LR_CREATEDIBSECTION);

	if(hbmp == NULL)
	{
		return false;
	}
	else
	{
		bool retVal = InterpretBits(hbmp); // interpret the bitmap data and return whether we succeeded
		DeleteObject(hbmp);
		return retVal;
	}
}

// can load JPEG and GIF files (can also load BMP, WMF, and TIF, if they are in 24 bit format) 
bool ImageExt::LoadJPGorGif(const char * filename, ImageX *img)
{
	m_pImg = img;
	strcpy(m_filename, filename);

	IPicture * picture;
	IStream * stream;
	HGLOBAL hGlobal;
	FILE * file;

	file = fopen(m_filename,"rb"); // open file in read only mode
	strcpy(last_filename, m_filename);
	if(file == NULL)
		return false;

	fseek(file,0,SEEK_END);
	int file_size = ftell(file);
	fseek(file,0,SEEK_SET);
	hGlobal = GlobalAlloc(GPTR, file_size); // allocates 112 k
	if(hGlobal == NULL)
	{
		fclose(file);
		return false;
	}
	fread((void*)hGlobal, 1, file_size, file);
	fclose(file);

	CreateStreamOnHGlobal(hGlobal,false,&stream);
	if(stream == NULL)
	{
		GlobalFree(hGlobal);
		return false;
	}

	// Decompress and load the JPG or GIF
	OleLoadPicture(stream,0,false,IID_IPicture,(void**)&picture); // 42,804 -> 43,844

	if(picture == NULL) { // allocates 1052 k
		stream->Release();
		GlobalFree(hGlobal);
		return false;
	}
	stream->Release();
	GlobalFree(hGlobal); // deallocates 52 k

	HBITMAP hB = 0;
	picture->get_Handle((unsigned int*)&hB);
	if(hB == 0) return false;

	HBITMAP hbmp = (HBITMAP)CopyImage(hB,IMAGE_BITMAP,0,0,LR_COPYRETURNORG);
	if(hbmp == 0) return false;

	picture->Release(); // deallocates 1024 k

	bool retVal = InterpretBits(hbmp);

	if(hB) DeleteObject(hB);
	if(hbmp) DeleteObject(hbmp);

	return retVal;
}

static bool doingPng = false;

bool ImageExt::InterpretBits(HBITMAP hbmp)
{
	HBITMAP hbmp2 = NULL;
	bool noMerge = false;
	bool ok = false;
	if(hbmp)
	{
		BITMAP bitmap;
		LPVOID bits;
		GetObject(hbmp,sizeof(bitmap),&bitmap);

		if(bitmap.bmBits != NULL)
		{
			// if the bits are conveniently where one would expect them to be (BMP)
			bits = bitmap.bmBits;
		} else
		{
			// if we have to do a little more work to locate the bits (JPG, GIF)
			int success = false;
			HDC hdc = CreateCompatibleDC(NULL);	// DC for Source Bitmap
			if(hdc)
			{
				HBITMAP hOldBitmap = (HBITMAP)SelectObject(hdc,hbmp);
				// here hdc contains the bitmap
					
				HDC hdc2 = CreateCompatibleDC(NULL); // DC for working
				if(hdc2)
				{
					// get bitmap size
					BITMAP bm;
					GetObject(hbmp, sizeof(bm), &bm);
							
					BITMAPINFO bitmapinfo; 
					ZeroMemory(&bitmapinfo, sizeof(BITMAPINFO));
					bitmapinfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
					bitmapinfo.bmiHeader.biWidth = bm.bmWidth;
					bitmapinfo.bmiHeader.biHeight = bm.bmHeight;
					bitmapinfo.bmiHeader.biPlanes = bm.bmPlanes;
					bitmapinfo.bmiHeader.biBitCount = 32;
					noMerge = true; // prevent from assuming 32-bit means alpha information
if(doingPng)
noMerge = false;
					UINT * pixels;	
					hbmp2 = CreateDIBSection(hdc2, (BITMAPINFO *)&bitmapinfo, 
													DIB_RGB_COLORS, (void **)&pixels, 
													NULL, 0);
					if(hbmp2)
					{
						HBITMAP hOldBitmap2 = (HBITMAP)SelectObject(hdc2, hbmp2);
						BitBlt(hdc2,0,0,
							   bm.bmWidth,bm.bmHeight,
							   hdc,0,0,SRCCOPY); // 42,768 -> 44,816
						SelectObject(hdc2, hOldBitmap2);
						success = true;
						bits = pixels; 
						bitmap.bmBitsPixel = bitmapinfo.bmiHeader.biBitCount;
						m_pImg->z_xres = bm.bmWidth;
						m_pImg->z_yres = bm.bmHeight;
						bitmap.bmWidthBytes = m_pImg->z_xres*bitmap.bmBitsPixel/8;
					}
				}
				SelectObject(hdc,hOldBitmap);
				if(hdc2) DeleteDC(hdc2); 
			}
			if(hdc) DeleteDC(hdc); 

			if(!success)
			{
///				if(bits) DeleteObject((HGDIOBJ) bits);
				return false;
			}
		}

/*		if(bitmap.bmWidth > 2048 || bitmap.bmHeight > 2048)
		{
			char errstr [256];
			sprintf(errstr,"The %dx%d image in file %s is too large for GameX to use.\nNeither width nor height may exceed 2048 pixels.",bitmap.bmWidth,bitmap.bmHeight,m_filename);
			MessageBox (NULL, errstr, "GameX Error", MB_OK|MB_ICONSTOP);
			if(bitmap.bmWidth > 2048) bitmap.bmWidth = 2048;
			if(bitmap.bmHeight > 2048) bitmap.bmHeight = 2048;
		}
*/
		int pitch = bitmap.bmWidthBytes;

		switch(bitmap.bmBitsPixel)
		{
		case 8:
		{
			HDC hdc = CreateCompatibleDC(NULL);
			HBITMAP hOldBitmap = (HBITMAP)SelectObject(hdc, hbmp);
			RGBQUAD palette[256];
			GetDIBColorTable(hdc, 0, 256, palette);
			m_pImg->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLOR);
			if(!m_pImg->z_data)
				return false;
			unsigned char * pData = (unsigned char *) m_pImg->z_data;
			unsigned char * bmpData = (unsigned char *) bits;
			for(int y = 0 ; y < m_pImg->z_yres ; y++)
			for(int x = 0 ; x < m_pImg->z_xres ; x++)
			{
				int n = (x+y*m_pImg->z_xres)*3; // pixel address -- destination
				int yi = m_pImg->z_yres-y-1; // inverted y, because bmp's are stored upside-down
				int ni = x+yi*pitch; // inverted pixel address -- source
				pData[n+2] = palette[bmpData[ni]].rgbBlue;   // blue
				pData[n+1] = palette[bmpData[ni]].rgbGreen; // green
				pData[n]   = palette[bmpData[ni]].rgbRed; // red
			}
			SelectObject(hdc, hOldBitmap);
			if(hdc) DeleteDC(hdc);
			break;
		}
		case 16:
		{
			int bmp_mask_red = 0x7C00, bmp_mask_green = 0x03E0, bmp_mask_blue = 0x001F;
			int bmp_shift_red, bmp_shift_green;
 
			bmp_shift_red   = GameX.z_GetShift(bmp_mask_red);
			bmp_shift_green = GameX.z_GetShift(bmp_mask_green);

			m_pImg->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_HIGHCOLOR);
			if(!m_pImg->z_data)
				return false;
			unsigned short * pData = (unsigned short *) m_pImg->z_data;
			unsigned short * bmpData = (unsigned short *) bits;
			for(int y = 0 ; y < m_pImg->z_yres ; y++)
			for(int x = 0 ; x < m_pImg->z_xres ; x++)
			{
				int n = x+y*m_pImg->z_xres; // pixel address -- destination
				int yi = m_pImg->z_yres-y-1; // inverted y, because bmp's are stored upside-down
				int ni = x+yi*pitch/2; // inverted pixel address -- source
				int clr = bmpData[ni] ;
				int blue  = clr & bmp_mask_blue;  // decode blue
				int green = (clr & bmp_mask_green) >> (bmp_shift_green-1);  // decode green
				int red   = (clr & bmp_mask_red) >> bmp_shift_red; // decode red
				pData[n] = ((red << 11) + (green << 5) + blue); // re-encode
			}
		 } break;
		case 24:
		{
			m_pImg->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLOR);
			if(!m_pImg->z_data)
				return false;
			unsigned char * pData = (unsigned char *) m_pImg->z_data;
			unsigned char * bmpData = (unsigned char *) bits;
			for(int y = 0 ; y < m_pImg->z_yres ; y++)
			for(int x = 0 ; x < 3*m_pImg->z_xres ; x += 3)
			{
				int n = x+y*m_pImg->z_xres*3; // pixel address -- destination
				int yi = m_pImg->z_yres-y-1; // inverted y, because bmp's are stored upside-down
				int ni = x+yi*pitch; // inverted pixel address -- source
				pData[n+2] = bmpData[ni];   // blue // (not only is it upside-down, but RGB is backwards)
				pData[n+1] = bmpData[ni+1]; // green
				pData[n]   = bmpData[ni+2]; // red
			}
		 } break;
		case 32:
		{
			if(noMerge) m_pImg->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLORX);
			else        m_pImg->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLORX|IMG_MERGEALPHA);
			if(!m_pImg->z_data)
				return false;
			unsigned char * pData = (unsigned char *) m_pImg->z_data;
			unsigned char * bmpData = (unsigned char *) bits;
			for(int y = 0 ; y < m_pImg->z_yres ; y++)
			for(int x = 0 ; x < 4*m_pImg->z_xres ; x += 4)
			{
				int n = x+y*m_pImg->z_xres*4; // pixel address -- destination
				int yi = m_pImg->z_yres-y-1; // inverted y, because bmp's are stored upside-down
				int ni = x+yi*pitch; // inverted pixel address -- source
				pData[n+3] = bmpData[ni+3]; // alpha
				pData[n+2] = bmpData[ni+2]; // red
				pData[n+1] = bmpData[ni+1]; // green
				pData[n]   = bmpData[ni]; // blue
			}


		 } break;
		}
		if(hbmp2) DeleteObject(hbmp2), hbmp2 = NULL;
		ok = true;
	}

	return ok;
}


bool ImageExt::SaveBmp(const char * filename, const HBITMAP hBitmap)
{
	bool ok = true;
	BITMAPINFO bmpInfo;
	ZeroMemory (&bmpInfo,sizeof(BITMAPINFO));
	bmpInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	HDC hdc = GetDC(NULL);
	GetDIBits (hdc, hBitmap, 0, 0, NULL, &bmpInfo, DIB_RGB_COLORS); 
	if(bmpInfo.bmiHeader.biSizeImage <= 0)
		bmpInfo.bmiHeader.biSizeImage = bmpInfo.bmiHeader.biWidth * abs(bmpInfo.bmiHeader.biHeight) * (bmpInfo.bmiHeader.biBitCount+7) / 8;
	LPVOID pixelBuffer = malloc(bmpInfo.bmiHeader.biSizeImage);
	if(pixelBuffer != NULL)
	{
		bmpInfo.bmiHeader.biCompression = BI_RGB;
		GetDIBits(hdc,hBitmap,0,bmpInfo.bmiHeader.biHeight,pixelBuffer, &bmpInfo, DIB_RGB_COLORS);
		FILE* file = fopen(filename,"wb");
		if(file != NULL)
		{
			BITMAPFILEHEADER bmpFileHeader;
			bmpFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER); 
			bmpFileHeader.bfReserved1 = 0;
			bmpFileHeader.bfReserved2 = 0;
			bmpFileHeader.bfSize = bmpFileHeader.bfOffBits + bmpInfo.bmiHeader.biSizeImage;
			bmpFileHeader.bfType = 'MB';
			fwrite(&bmpFileHeader,sizeof(BITMAPFILEHEADER),1,file);
			fwrite(&bmpInfo.bmiHeader,sizeof(BITMAPINFOHEADER),1,file);
			fwrite(pixelBuffer,bmpInfo.bmiHeader.biSizeImage,1,file); 
		}
		else
		{
			debug.Output("ImageExt::SaveBmp: Unable to Create Bitmap File");
			ok = false;
		}
		fclose(file);
	} else
	{
		debug.Output("ImageExt::SaveBmp: Unable to Allocate Bitmap Memory");
		ok = false;
	}
	if(hdc) ReleaseDC(NULL,hdc); 
	if(pixelBuffer) free(pixelBuffer); 
	return ok;
}





/* The helper function to read the PNG's bits from file
   makes use of the Open Source-licensed libpng and zLib. 
   see png.h for more details. */

#define PNG_NO_WRITE_SUPPORTED
#define PNG_NO_CONSOLE_IO
///#define PNG_NO_STDIO

#include "png.h"

/* Read a PNG file. You may want to return an error code if the read
* fails (depending upon the failure). There are two "prototypes" given
* here - one where we are given the filename, and we need to open the
* file, and the other where we are given an open file (possibly with
* some or all of the magic bytes read - see comments above).
*/
boolean ReadPNG(
	const char * filename,
	int nLen,
	LPBYTE& lpBuffer,
	int& width,
	int& height,
	boolean& bHasAlpha )
{
	png_structp png_ptr;
	png_infop info_ptr;
	FILE *fp = NULL;

	if(nLen == 0)
	{
		if((fp = fopen(filename, "rb")) == NULL)
			return false;
	}

	/* Create and initialize the png_struct with the desired error handler
	* functions. If you want to use the default stderr and longjump method,
	* you can supply NULL for the last three parameters. We also supply the
	* the compiler header file version, so that we know if the application
	* was compiled with a compatible version of the library. REQUIRED
	*/
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL);

	if(png_ptr == NULL)
	{
		fclose(fp);
		return false;
	}

	/* Allocate/initialize the memory for image information. REQUIRED. */
	info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr == NULL)
	{
		fclose(fp);
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return false;
	}

	/* Set error handling if you are using the setjmp/longjmp method (this is
	* the normal method of doing things with libpng). REQUIRED unless you
	* set up your own error handlers in the png_create_read_struct() earlier.
	*/

	if(setjmp(png_jmpbuf(png_ptr)))
	{
		/* Free all of the memory associated with the png_ptr and info_ptr */
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(fp);
		/* If we get here, we had a problem reading the file */
		return false;
	}

	/* One of the following I/O initialization methods is REQUIRED */
	//if(nLen == 0)
	{
		/* Set up the input control if you are using standard C streams */
		png_init_io(png_ptr, fp);
	}
	//else
	//{
	//	/* If you are using replacement read functions, instead of calling
	//	* png_init_io() here you would call:
	//	*/
	//	png_set_read_fn(png_ptr, (void *)filename, user_read_fn);
	//	/* where user_io_ptr is a structure you want available to the callbacks */
	//}

	/* If we have already read some of the signature */
	png_set_sig_bytes(png_ptr, 0);

	/*
	* If you have enough memory to read in the entire image at once,
	* and you need to specify only transforms that can be controlled
	* with one of the PNG_TRANSFORM_* bits (this presently excludes
	* dithering, filling, setting background, and doing gamma
	* adjustment), then you can read the entire image (including
	* pixels) into the info structure with this call:
	*/
	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_EXPAND|PNG_TRANSFORM_BGR|PNG_TRANSFORM_STRIP_16, NULL);

	int padding = 0; // Padding at end of each row
	lpBuffer = new BYTE[info_ptr->width*info_ptr->height*4];
	width = info_ptr->width;
	height = info_ptr->height;

	png_bytepp rows = png_get_rows(png_ptr, info_ptr);

	byte* pDest = lpBuffer;
	byte* pSrc;
	byte b; // Used for gray
	if(info_ptr->channels == 1)// gray
	{
		// rows in RGB
		bHasAlpha = false;
		for(UINT r = 0; r < info_ptr->height; r++)
		{
			pSrc = rows[r];
			for(int c = info_ptr->width; c-- > 0; )
			{
				*pDest++ = 0xff;
				b = *pSrc++;
				*pDest++ = b;
				*pDest++ = b;
				*pDest++ = b;
			}
			pDest += padding;
		}
	}
	else if(info_ptr->channels == 2)// gray
	{
		// rows in RGB
		bHasAlpha = false;
		for(UINT r = 0; r < info_ptr->height; r++)
		{
			pSrc = rows[r];
			for(int c = info_ptr->width; c-- > 0; )
			{
				*pDest++ = *pSrc++;
				b = *pSrc++;
				*pDest++ = b;
				*pDest++ = b;
				*pDest++ = b;
			}
			pDest += padding;
		}
	}
	else if(info_ptr->channels == 3)
	{
		// rows in RGB
		bHasAlpha = false;
		for(UINT r = 0; r < info_ptr->height; r++)
		{
			pSrc = rows[r];
			for(int c = info_ptr->width; c-- > 0; )
			{
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
				*pDest++ = *pSrc++;
				*pDest++ = 0xff;
			}
			pDest += padding;
		}
	}
	else if(info_ptr->channels == 4)
	{
		bHasAlpha = true;
		for(UINT r = 0; r < info_ptr->height; r++)
		{
			pSrc = rows[r];

			memcpy(pDest, pSrc, info_ptr->width*4);
			pDest += info_ptr->width*4;
			pDest += padding;
		}
	}


	/* At this point you have read the entire image */

	/* clean up after the read, and free any memory allocated - REQUIRED */
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	/* close the file */
	fclose(fp);

	/* that's it */
	return true;
}


// loads a JPG image from file into an ImageX
bool ImageExt::LoadPNG (const char * filename, ImageX *img)
{
	m_pImg = img;
	strcpy (m_filename, filename);
	strcpy (last_filename, filename);


	int wd = 64, ht = 64;
	LPBYTE pBits = NULL;

	boolean bHasAlpha = TRUE;
	ReadPNG( filename, 0, pBits, wd, ht, bHasAlpha );
	if(!pBits)
		return false;

	int pitch = wd * 4;
	m_pImg->Size(wd, ht, IMG_TRUECOLORX|(bHasAlpha ? IMG_MERGEALPHA : 0));
	unsigned char * pData = (unsigned char *) m_pImg->z_data;
	unsigned char * bmpData = (unsigned char *) pBits;
	for(int y = 0 ; y < m_pImg->z_yres ; y++)
	for(int x = 0 ; x < 4*m_pImg->z_xres ; x += 4)
	{
		int n = x+y*m_pImg->z_xres*4; // pixel address -- destination
		int ni = x+y*pitch; // inverted pixel address -- source
		pData[n+3] = bmpData[ni+3]; // alpha
		pData[n+2] = bmpData[ni+2]; // red
		pData[n+1] = bmpData[ni+1]; // green
		pData[n]   = bmpData[ni]; // blue
	}

	SAFE_DELETE_ARRAY(pBits);

	return true;

}







