//
// GameX - Initialization- and Shutdown-related GameXClass Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// R5G5B5 display mode (sometimes called 15-bit mode) may not be supported,
// because I've never actually seen a monitor that uses it.
// i.e., R5G5B5 mode *should* work but has not been tested, so it should be tested sometime soon.

#include "gamex-win-dx.h"
#include "gamex-utilities.h"
#include "gamex-win-dx-timer.inl"
#include "gamex-image.h"
#include "gamex-sound.h"
#include "gamex-text.h"
#include "gamex-camera.h"

// Direct3D
#ifndef DIRECT3D_VERSION
#define DIRECT3D_VERSION 0x0700 // specify version 7.0 of Direct3D -- the last version that's compatible with DirectDraw
#include <d3d.h>				// note: for newer versions of D3D we would need to 
//#include <d3dtypes.h>			//		 include different files, like d3d8.h
//#include "d3dx.h"
#endif

// DirectMusic
#ifndef DIRECTMUSIC_VERSION
#define DIRECTMUSIC_VERSION 0x0700 // specify version 7.0 of DirectMusic
#include <dmksctrl.h>
#include <dmusici.h>
#include <dmusicc.h>
#include <dmusicf.h>
#endif

// DirectShow
#ifndef DIRECTSHOW_VERSION
#define DIRECTSHOW_VERSION 0x0700 // specify version 7.0 of DirectShow
#include <strmif.h>
#include <control.h>
#include <uuids.h>
#include <evcode.h>
#endif


// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif


#ifndef SAFE_RELEASE
	#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p) = NULL; } }
#endif
#ifndef SAFE_UNACQUIRE_RELEASE
	#define SAFE_UNACQUIRE_RELEASE(p)      { if(p) { (p)->Unacquire(); (p)->Release(); (p) = NULL; } }
#endif

CameraX * defaultCam = NULL;

typedef DECLSPEC_IMPORT UINT (WINAPI* LPTIMEBEGINPERIOD) (UINT uPeriod);

bool z_GameX_exists = false;

// GameXClass Constructor
GameXClass::GameXClass ()
{
	// enforce this class being a singleton
	{
		static bool hasInstance = false;
		if(hasInstance)
		{
			MessageBox(NULL, "ERROR: GameX must be a singleton object, but more than 1 was detected.", "GameX Error",MB_OK|MB_ICONERROR);
			exit(-1);
		}
		hasInstance = true;
	}

	z_GameX_exists = true;

	BeginCheckingUnfreed();

	z_win_request_xsize = -1;
	z_win_request_ysize = -1;
	z_win_status = GAMEX_NO_INIT;
	z_win_init_done = false;
///	z_sound_list = NULL;
	z_sound_support = false;
	z_sound_support_tested = false;
	z_drawLocked = false;
	z_win_warning_on_draw = false;
	strcpy(z_win_name, "GameX");
	z_draw_state = new GameXDrawState();
	defaultCam = new CameraX();
	z_draw_state->cam = defaultCam;
	z_draw_state->dst = VIEWPORT;
	ResetDrawStates();
	z_d3d_vbuffer = NULL;
	z_d3d_vbuffermem = NULL;
	z_d3d_vbufferfilled = 0;
	z_d3d_vbufferlocked = false;

	if(debug.IsRunningDebug())
	{
		z_win_log_debug = true;
		z_win_err_debug = true;
	}
	else
	{
		z_win_log_debug = false;
		z_win_err_debug = true;
	}

	for(int i = 0 ; i < GAMEX_TOTAL_STATUS_NUMBER ; i++)
		z_win_status_problem_ignore[i] = false;
	z_win_appIconIDStr = IDI_APPLICATION;

	z_defaultCursorImg = NULL;

	z_di_mousestate = new DIMOUSESTATE;
	z_d3d_zbuffer_format = new DDPIXELFORMAT;

	z_dm_loader = NULL;
	z_dm_performance = NULL;
	z_mp3_support_tested = false;

	z_ds_main = NULL;
	z_master_music_tempo = z_master_music_volume = 1.0f;

	z_d3d_vbufferenabled = false;
	z_dd_back_buf = NULL;

	z_win_cursor_hidden = false;
	z_lastPercent = 0;

	RandomizeSeed();

/*
	extern WORD triangleListIndex [D3DMAXNUMVERTICES*6/4];
	int count = 0;
	int j = -2;
	for(int i = 0 ; i < D3DMAXNUMVERTICES*6/4 ; i++)
	{
		if((count-4)%6 == 0)
			j -= 2;
		if((count-5)%6 == 0)
			j -= 2;
		if((count-6)%6 == 0)
			j += 2;

		triangleListIndex[i] = j;
		j++;
		count++;
	}
*/

    // Increase the accuracy of Sleep() without needing to link to winmm.lib
    CHAR pszPath[MAX_PATH+1];
    if(GetSystemDirectory(pszPath, MAX_PATH+1))
	{
		lstrcat(pszPath, "\\winmm.dll");
		HINSTANCE hInstWinMM = LoadLibrary(pszPath);
		if(hInstWinMM != NULL) 
		{
			LPTIMEBEGINPERIOD pTimeBeginPeriod = (LPTIMEBEGINPERIOD)GetProcAddress(hInstWinMM, "timeBeginPeriod");
			if(NULL != pTimeBeginPeriod)
				pTimeBeginPeriod(1);
		}
		FreeLibrary(hInstWinMM);
	}


	// make sure the linker tries referencing this symbol so the "forgot-to-define-GAMEX_MAIN" error makes sense
	GAMEX_MAIN_MUST_BE_DEFINED_IN_YOUR_MAIN_CPP_FILE_WHICH_MUST_DIRECTLY_INCLUDE_GAMEX_H._ = 0;
}

///static HANDLE freezeCheckThread = NULL;

GameXClass::~GameXClass ()
{
	if(z_win_log_debug)
		debug.Output("Uninitializing GameX...\n");

///	if(freezeCheckThread)
///		TerminateThread(freezeCheckThread, 0);

	SAFE_DELETE(z_draw_state);
	SAFE_DELETE(defaultCam);

	SAFE_DELETE(z_defaultCursorImg);
	SAFE_DELETE(z_di_mousestate);
	SAFE_DELETE(z_d3d_zbuffer_format);

	// do the main uninitializing of GameX
	z_CleanUp();

	if(z_win_log_debug)
	{
		debug.Output("Exiting...\n");
		debug.OutputTime();
	}
	debug.Flush();

#ifdef _DEBUG // some internal verification if debugging GameX
	GameX.z_LoadQueuedImages();
	assertNoLib(z_queued_image_loads.size() == 0);
	assertNoLib(z_copying_images.size() == 0);
	assertNoLib(z_copyable_images_loaded.size() == 0);
#endif

	z_GameX_exists = false;
}


// This function must be called by the game's GameInit() function
// Parameters: Name, Flags, ViewWidth, ViewHeight, [optional:]	Speed, Smoothness, SkipLimit 
bool GameXClass::Initialize (const char * name, GameXInitFlags options, int xres, int yres, int speed)
{
	strcpy(z_win_name,name); // remember game name

	z_d3d_vbufferenabled = false;

	if(z_win_log_debug)
	{
		if(options & VIDEO_WINDOWED && options & VIDEO_FULLSCREEN)
		{
			MessageBox(z_win_hwnd, "WARNING: VIDEO_WINDOWED and VIDEO_FULLSCREEN are mutually exclusive flags.", "GameX Warning",MB_OK|MB_ICONWARNING);
			options &= ~(VIDEO_WINDOWED | VIDEO_FULLSCREEN);
		}
		if(options & VIDEO_ALLOWREFRESHSYNC && options & VIDEO_DISALLOWREFRESHSYNC)
		{
			MessageBox(z_win_hwnd, "WARNING: VIDEO_ALLOWREFRESHSYNC and VIDEO_DISALLOWREFRESHSYNC are mutually exclusive flags.", "GameX Warning",MB_OK|MB_ICONWARNING);
			options &= ~(VIDEO_ALLOWREFRESHSYNC | VIDEO_DISALLOWREFRESHSYNC);
		}
		if(options & VIDEO_SOFTFULLSCREEN && options & VIDEO_HARDFULLSCREEN)
		{
			MessageBox(z_win_hwnd, "WARNING: VIDEO_SOFTFULLSCREEN and VIDEO_HARDFULLSCREEN are mutually exclusive flags.", "GameX Warning",MB_OK|MB_ICONWARNING);
			options &= ~(VIDEO_SOFTFULLSCREEN | VIDEO_HARDFULLSCREEN);
		}
	}


	z_quit_dialog = false;

	if(options & VIDEO_ASKFULLSCREEN)
		z_win_request_full = 2;
	else if(options & VIDEO_WINDOWED)
		z_win_request_full = 0;
	else if(options & VIDEO_FULLSCREEN)
		z_win_request_full = 1;
	else
		z_win_request_full = 3;

	if(options & VIDEO_NORESIZE)
		z_win_request_can_resize = false;
	else
		z_win_request_can_resize = true;

	if(options & VIDEO_LOWRES)
		z_win_request_low_resolution = true;
	else
		z_win_request_low_resolution = false;

	if(options & VIDEO_SOFTFULLSCREEN)
		z_win_request_soft_fullscreen = true;
	else
		z_win_request_soft_fullscreen = false;

	if(options & VIDEO_HARDFULLSCREEN)
		z_win_request_hard_fullscreen = true;
	else
		z_win_request_hard_fullscreen = false;

	if(xres == 0 && yres == 0)
	{
		z_win_request_full = 1;
		z_win_request_soft_fullscreen = true;
	}

	z_win_fake_fullscreen = z_win_request_soft_fullscreen ? true : (z_win_request_hard_fullscreen ? false : debug.IsRunningDebug());

	if(z_win_request_full == 2)
	{
		int style = MB_YESNOCANCEL|MB_ICONQUESTION;

		if(options & VIDEO_FULLSCREEN || !(options & VIDEO_WINDOWED) && !debug.IsRunningDebug())
			style |= MB_DEFBUTTON1;
		else //if(options & VIDEO_WINDOWED || !(options & VIDEO_FULLSCREEN) && debug.IsRunningDebug())
			style |= MB_DEFBUTTON2;

		int choice = MessageBox(NULL, "Do you want to play this game in fullscreen mode?", z_win_name, style);
		if(choice == IDYES)
			z_win_request_full = 1;
		else if(choice == IDNO)
			z_win_request_full = 0;
		else if(choice == IDCANCEL)
		{
			debug.Flush();
			TerminateProcess(GetCurrentProcess(), 0); // extreme prejudice
		}
	}
	else if(z_win_request_full == 3)
	{
		if(debug.IsRunningDebug())
			z_win_request_full = 0;
		else
			z_win_request_full = 1;
	}

	if(options & VIDEO_DISALLOWREFRESHSYNC)
		z_win_request_vsync = 0;
	else
		z_win_request_vsync = 1;

	if(options & VIDEO_SOFTWAREONLY)
	{
		MessageBox (z_win_hwnd, "WARNING: Entering software-only mode because VIDEO_SOFTWAREONLY was turned on.\nThis is just a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION);
		z_win_request_software_only = true;
	}
	else
		z_win_request_software_only = false;

	if(options & VIDEO_SLOWDEBUGDRAWING)
	{
		MessageBox (z_win_hwnd, "WARNING: Entering ultra-slow drawing mode because the VIDEO_SLOWDEBUGDRAWING was set.\nJust a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION);
		z_win_request_slow_debug_drawing = true;
//		speed = 5;
	}
	else
		z_win_request_slow_debug_drawing = false;

	if(options & RUN_KEEPCURSORINWINDOW)
		z_win_request_keep_cursor_in_window = true;
	else
		z_win_request_keep_cursor_in_window = false;

	z_win_request_delete_mdata = true;

	z_win_request_bpp = 0;
	if(options & VIDEO_16BIT)
	{
		z_win_request_bpp = 16; // 16 bit color is fully supported by GameX
	}
	else if(options & VIDEO_24BIT)
	{
		assertMsg(0, "24-bit RGB color mode is not supported - using 32-bit instead...");
		z_win_request_bpp = 32;
	}
	else if(options & VIDEO_32BIT)
	{
		z_win_request_bpp = 32; // 32 bit color is fully supported by GameX -- may be a little slow though
	}

	if(z_win_log_debug)
	{
		if(1 < /*!!(options & VIDEO_8BIT)+*/!!(options & VIDEO_16BIT)+!!(options & VIDEO_24BIT)+!!(options & VIDEO_32BIT))
		{
			debug.Flush();
			MessageBox (z_win_hwnd, "ERROR: You cannot force select more than one video mode at once.\nPlease initialize GameX with only one VIDEO_??BIT flag, or with none if any video mode is OK.", "GameX Error",MB_OK|MB_ICONSTOP);
			options &= ~(VIDEO_16BIT | VIDEO_24BIT | VIDEO_32BIT);
		}
	}

	z_win_request_xsize = xres;
	z_win_request_ysize = yres;

	z_win_cps = speed; // see WinGameRun() for usage

	if(z_win_cps != 0) // if it isn't set to unmanaged
	if(z_win_cps < 5 || z_win_cps > 600)
	{
		if(z_win_log_debug)
		{
			char disp [500] ;
			sprintf (disp, "WARNING: You have specified %d CPS, which is invalid. The CPS must be at least 5, at most 600, and normally should be between 30 and 120.", z_win_cps);
			debug.Flush();
			MessageBox (z_win_hwnd, disp, "GameX Warning",MB_OK|MB_ICONWARNING);
		}
		if(z_win_cps < 5) z_win_cps = 5;
		else if(z_win_cps > 600) z_win_cps = 600;
	}

	z_win_fps = -1; // default of targeting the best possible framerate that's visible
	z_win_max_cpf = max(3,z_win_cps/5); // default to skip no more than 1/5 of a second of frames per instant


	if(options & AUDIO_DISABLE)
		z_win_request_no_sound = true;
	else
		z_win_request_no_sound = false;

	if(options & RUN_USEESCAPEKEY)
		z_win_request_use_escape = true;
	else
		z_win_request_use_escape = false;

	if(options & RUN_NOALTENTERTOGGLE)
		z_win_request_use_alt_enter = true;
	else
		z_win_request_use_alt_enter = false;

	if((options & RUN_AUTOSHOWINFO) || debug.IsRunningDebug())
		z_win_request_show_run_info = true;
	else
		z_win_request_show_run_info = false;

	if(options & RUN_BACKGROUND)
		z_win_request_no_auto_pause = true;
	else
		z_win_request_no_auto_pause = false;

	if(options & RUN_ALWAYSKEEPCURSOR && options & RUN_ALWAYSHIDECURSOR)
	{
		if(z_win_log_debug)
			MessageBox (z_win_hwnd, "WARNING: RUN_ALWAYSKEEPCURSOR and RUN_ALWAYSHIDECURSOR are mutually exclusive flags.", "GameX Warning",MB_OK|MB_ICONWARNING);
		options ^= RUN_ALWAYSKEEPCURSOR;
		options ^= RUN_ALWAYSHIDECURSOR;
	}

	if(options & RUN_ALWAYSKEEPCURSOR)
		z_win_request_no_hide_cursor = true;
	else
		z_win_request_no_hide_cursor = false;

	if(options & RUN_ALWAYSHIDECURSOR)
		z_win_request_no_show_cursor = true;
	else
		z_win_request_no_show_cursor = false;

	if(options & RUN_NOMOUSEINPUT)
		z_win_request_no_mouse_input = true;
	else
		z_win_request_no_mouse_input = false;

	if((options & RUN_NOCONFIRMQUIT) || debug.IsRunningDebug())
		z_win_request_no_confirm_quit = true;
	else
		z_win_request_no_confirm_quit = false;

	bool success = true;

	if(z_win_log_debug)
		debug.Output("GameX: Initialization started.\n");

	if(z_InitWindowSetup() == false) success = false;
	if(success)
	if(z_InitDirectX() == false) success = false;
	
	if(!success)
	{
		z_ReportProblem(z_GetStatus());
		if(z_win_err_debug)
			debug.Output("GameX: Initialization had problems.\n");
		return false;
	}

	z_DoInitTests();

	z_DisplayLoadingWithValue (-5, "Loading Text...");

	// initialize the default cursor
	if(!z_win_request_no_show_cursor)
	{
		z_defaultCursorImg = new ImageX();
		z_defaultCursorImg->Create(16,32,false); 
		z_defaultCursorImg->UseWith(z_defaultCursorImg->GetUsage() | LOAD_MASKED);
		const int width = 11;
		const int height = 19;

		// a little bit of hard-coding here...
		unsigned char defaultCursorBits [height][width] =
		{
			0x8,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
			128,0x8,  0,  0,  0,  0,  0,  0,  0,  0,  0,
			128,192,0x8,  0,  0,  0,  0,  0,  0,  0,  0,
			128,192,192,0x8,  0,  0,  0,  0,  0,  0,  0,
			128,255,192,192,0x8,  0,  0,  0,  0,  0,  0,
			128,255,192,192,192,0x8,  0,  0,  0,  0,  0,
			128,255,255,192,192,192,0x8,  0,  0,  0,  0,
			128,255,255,192,192,192,192,0x8,  0,  0,  0,
			128,255,255,255,192,192,192,192,0x8,  0,  0,
			128,255,255,255,192,192,192,192,192,0x8,  0,
			128,255,255,192,0x8,0x8,0x8,0x8,0x8,0x8,0x8,
			128,255,192,0x8,255,192,0x8,  0,  0,  0,  0,
			128,192,0x8,  0,128,255,192,0x8,  0,  0,  0,
			128,0x8,  0,  0,128,255,192,0x8,  0,  0,  0,
			0x8,  0,  0,  0,  0,128,255,192,0x8,  0,  0,
			  0,  0,  0,  0,  0,128,255,192,0x8,  0,  0,
			  0,  0,  0,  0,  0,  0,128,255,192,0x8,  0,
			  0,  0,  0,  0,  0,  0,128,192,192,0x8,  0,
			  0,  0,  0,  0,  0,  0,  0,0x8,0x8,0x8,  0,
		};
		AccessPixels(z_defaultCursorImg);
		for(int y = 0 ; y < height ; y++)
		for(int x = 0 ; x < width ; x++)
		{
			const int val = defaultCursorBits[y][x];
			DrawPixel(x,y, val,val,val);
		}
		EndPixelAccess();
	}

	z_DisplayLoadingWithValue (-10, "Loading Game...");

	if(z_win_log_debug)
		debug.Output("GameX: Initialization finished OK.\n");

	///z_d3d_vbufferenabled = true;
	z_d3d_vbufferenabled = false; // default

//	SetForegroundWindow(z_win_hwnd);

	return true;
}

void GameXClass::z_DoInitTests ()
{
	z_win_any_draw_mode_workaround = 0;

	z_win_status = GAMEX_CREATING_TEST_SURFACE;

	// test of blurring speed:
	z_win_blur_supported = true; // temporarily true to allow testing
	int trials = 10;
	long long timer = GetClock();
	for(int i = 0 ; i < trials ; i++)
	{
		BlurScreen(0.51f);
	}
	timer = GetClock() - timer;
	double testTime = (double)(timer) / (double)GetClocksPerSec() / (double)trials;
	double minTime = 0.025*(double)(z_win_width*z_win_height)/(double)(800*600); // minimum acceptable time for test (5x slower than decent speed of 0.005 seconds/trial/(800*600)pixels)
	if(testTime > minTime)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Blur speed test FAILED, blurring disabled.\n");
		z_win_blur_supported = false;
	}
	else
	{
		z_win_blur_supported = true;
	}



	// fill the background with 50% gray

	// (make sure it worked, while we're at it:)
	int r = 0,g = 0,b = 0;
	int tol = 24; // tolerance allowed in each direction from expected color component -- necessary because of color format imprecision as well as slight deviations in video card processing results

	// create test images:
	ImageX testImage, testImage2, testAlphaImage, testDDImage;
	testImage.Create(16,16);
	testImage.Fill(255,0,0); // fill with bright red
	testImage2.Create(16,16);
	testImage2.Fill(192,128,64); // fill with brownish color
	testAlphaImage.Create(16,16,true);
	testAlphaImage.Fill(255,0,0,255); // fill with bright red (full opacity)
	testDDImage.Create(64,64);
	testDDImage.ConvertUsageTo(LOAD_NON3D);
	testDDImage.Fill(255,0,0); // fill with bright red

	// Draw Mode Tests:
	// Some video cards handle certain drawing modes improperly,
	// so we test for correctness here and enable work-arounds as necessary:

	// Subtractive (DRAW_SUBTRACT mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_subtract_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_SUBTRACT);
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (0,128,128) after this line
	EndPixelAccess();
	if(r>0+tol || r<0-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol)
	{
		if(z_win_fullscreen) ToggleFullscreen();
		
		FillScreen(ColorX(128,128,128));
		SetDrawMode(DRAW_SUBTRACT);
		DrawImage(&testAlphaImage, 0,0);

		AccessPixels(VIEWPORT);
		ReadPixel(0,0,r,g,b); // (r,g,b) should equal (0,128,128) after this line
		EndPixelAccess();

		if(r>0+tol || r<0-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol)
		{
			if(z_win_err_debug)
				debug.Output("GameX: Subtractive drawing test FAILED, black-alpha work-around enabled.\n");
			z_win_subtract_workaround = 2; // black-alpha subtraction simulation required (worst case)  uses image with all black and varying alpha image to darken destination non-subtractively
		}
		else
		{
			if(z_win_err_debug)
				debug.Output("GameX: Subtractive drawing test FAILED, 32-bit work-around enabled.\n");
			z_win_subtract_workaround = 1; // subtractive drawing works with alpha images although not normally, so enable workaround that converts images to alpha if drawn subtractively
			z_win_any_draw_mode_workaround = true;
		}
		
	}
	else
	{
		z_win_subtract_workaround = 0; // no workaround required
	}

	// Multiplicatively Additive (DRAW_INTENSIFY mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_intensify_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_INTENSIFY);
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (255,128,128) after this line
	EndPixelAccess();
	if(r>255+tol || r<255-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Intensify drawing test FAILED, additive work-around enabled.\n");
		z_win_intensify_workaround = 1; // workaround: simply use regular addition instead of multiplicative addition
		z_win_any_draw_mode_workaround = true;
	}
	else
	{
		z_win_intensify_workaround = 0; // no workaround required
	}

	// Color-valued Alpha (DRAW_GHOST mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_ghost_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_GHOST);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (192,128,112) after this line
	EndPixelAccess();
	if(r>192+tol || r<192-tol || g>128+tol || g<128-tol || b>112+tol || b<112-tol)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Ghost drawing test FAILED, alpha work-around enabled.\n");
		z_win_ghost_workaround = 1; // workaround required
		z_win_any_draw_mode_workaround = true;
	}
	else
	{
		z_win_ghost_workaround = 0; // no workaround required
	}

	// DRAW_BURN Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_burn_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_BURN);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (64,128,144) after this line
	EndPixelAccess();
	if(r>64+tol || r<64-tol || g>128+tol || g<128-tol || b>144+tol || b<144-tol)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Burn drawing test FAILED, invert+subtract work-around enabled.\n");
		z_win_burn_workaround = 1; // workaround required
		z_win_any_draw_mode_workaround = true;
	}
	else
	{
		z_win_burn_workaround = 0; // no workaround required
	}

	
	// DRAW_ADDSOFT Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_addsoft_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_ADDSOFT);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (224,192,160) after this line
	EndPixelAccess();
	if(r>224+tol || r<224-tol || g>192+tol || g<192-tol || b>160+tol || b<160-tol)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Soft additive drawing test FAILED, regular addition work-around enabled.\n");
		z_win_addsoft_workaround = 1; // workaround: simply use regular addition instead
		z_win_any_draw_mode_workaround = true;
	}
	else
	{
		z_win_addsoft_workaround = 0; // no workaround required
	}


	// DRAW_ADDSHARP Drawing Test:

	FillScreen(ColorX(0,0,0)); // (on top of a black background)

	z_win_addsharp_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_ADDSHARP);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (255,255,64) after this line
	EndPixelAccess();
	if(r>255+tol || r<255-tol || g>255+tol || g<255-tol || b>64+tol || b<64-tol)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Sharp additive drawing test FAILED, regular addition work-around enabled.\n");
		z_win_addsharp_workaround = 1; // workaround: simply use regular addition instead
		z_win_any_draw_mode_workaround = true;
	}
	else
	{
		z_win_addsharp_workaround = 0; // no workaround required
	}


	// Alpha-blending Test:
	// (some video cards don't apply alpha blending properly except on alpha images or non-textured polygons)
	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	z_win_alpha_blend_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_PLAIN);
	SetDrawShading(ColorX(255,255,255,128));
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should about equal (192,64,64) after this line
	EndPixelAccess();
	if(r>192+tol || r<192-tol || g>64+tol || g<64-tol || b>64+tol || b<64-tol)
	{
		FillScreen(ColorX(128,128,128));
		SetDrawMode(DRAW_PLAIN);
		SetDrawShading(ColorX(255,255,255,128));
		DrawImage(&testAlphaImage, 0,0);

		AccessPixels(VIEWPORT);
		ReadPixel(0,0,r,g,b); // (r,g,b) should about equal (192,64,64) after this line
		EndPixelAccess();

		if(r>192+tol || r<192-tol || g>64+tol || g<64-tol || b>64+tol || b<64-tol)
		{
			if(z_win_err_debug)
				debug.Output("GameX: Alpha-blend shading test FAILED, some alpha shading disabled.\n");
			z_win_alpha_blend_workaround = 0; // no workaround known in this case
		}
		else
		{
			if(z_win_err_debug)
				debug.Output("GameX: Alpha-blend shading test FAILED, 32-bit work-around enabled.\n");
			z_win_alpha_blend_workaround = 1; // if works on alpha images, convert images to 32-bit if drawn with alpha shading
			z_win_any_draw_mode_workaround = true;
		}
	}
	else
	{
		z_win_alpha_blend_workaround = 0; // no workaround required
	}

	long double nonStretchTime, stretchTime;
	{
		int trials = 50;
//		for(int i = 0 ; i < trials/50 ; i++)
//			DrawTexturedRect(&testDDImage, 0,0,64,64);
	}
	{
		// test of non-stretching speed:
		int trials = 25;
		long long timer = GetClock();
		for(int i = 0 ; i < trials ; i++)
		{
			DrawTexturedRect(&testDDImage, 0,0,64,64);
		} 
		timer = GetClock() - timer;
		nonStretchTime = 1000.0 * (long double)(timer) / (long double)GetClocksPerSec() / (long double)trials;
	}
	{
		// test of stretching speed:
		int swidth = min(400,GameX.GetWidth());
		int sheight = min(400,GameX.GetHeight());
		int trials = 25;
		long long timer = GetClock();
		for(int i = 0 ; i < trials ; i++)
		{
			DrawTexturedRect(&testDDImage, 0,0,(float)swidth,(float)sheight);
		}
		timer = GetClock() - timer;
		stretchTime = 1000.0 * (long double)(timer) / (long double)GetClocksPerSec() / (long double)trials;
	}

	ClearScreen();
	z_win_status = GAMEX_READY;
	UpdateView();

	if(stretchTime > nonStretchTime*10.0f)
	{
		if(z_win_err_debug)
			debug.Output("GameX: Stretched blitting was too slow (%g times slower than non-stretched), non-stretch workaround enabled.\n", (double)(stretchTime / nonStretchTime));

		z_win_resize_workaround = 1;

		RECT rect; // in case window began as too large already:
		GetClientRect (z_win_hwnd, &rect);
		z_win_xoffset = (rect.right-rect.left - z_win_width) / 2;
		z_win_yoffset = (rect.bottom-rect.top - z_win_height) / 2;				
		z_win_realx = z_win_width;
		z_win_realy = z_win_height;

		// keep the cursor inside the game if we're in fullscreen or were requested to keep it in the window
		if(z_win_fullscreen || z_win_request_keep_cursor_in_window)
		{
			RECT rect = {z_win_xpos+z_win_xoffset, z_win_ypos+z_win_yoffset, z_win_xpos+z_win_xoffset+z_win_realx, z_win_ypos+z_win_yoffset+z_win_realy} ;
			ClipCursor(&rect);
		}
		else
		{
			ClipCursor(NULL);
		}
	}
	else
	{
		if(z_win_log_debug)
			debug.Output("GameX: Stretched blitting was acceptably fast (%g times slower than non-stretched).\n", (double)(stretchTime / nonStretchTime));
	}

}

bool GameXClass::z_InitWindowSetup ()
{
	if(z_InitWindowClass() == false)		return false;
	if(z_InitWindows() == false)			return false;
	if(z_WindowDraw() == false)			return false;
//	SendMessage(z_win_hwnd,WM_SETICON,ICON_SMALL,-1); // remove icon since I can't seem to get it to display the one it should
	return true;
}

bool GameXClass::z_InitDirectX ()
{	
	if(FAILED(CoInitialize(NULL)))
	{
		z_win_status = GAMEX_DDSTART_FAILED;
		return false;
	}

	bool success = true;

///	if(GetAsyncKeyState(VK_ESCAPE) & 0x8000)
///		Quit(true);

	if(z_InitDirectDraw() == false)
	{
		success = false;
		z_ReportProblem();
		if(z_win_err_debug)
			debug.Output("GameX: DirectDraw init failed.\n");
	} else
	{
		if(z_win_log_debug)
			debug.Output("GameX: DirectDraw init succeeded.\n");
	}

	z_DisplayLoadingWithValue (-5, "Initializing Direct3D...");

	if(z_InitDirect3D() == false)
	{
		success = false;
		z_ReportProblem();
		if(z_win_err_debug)
			debug.Output("GameX: Direct3D init failed.\n");
	}
	else
	{
		if(z_win_log_debug)
			debug.Output("GameX: Direct3D init succeeded.\n");
	}

	z_DisplayLoadingWithValue (-10, "Initializing DirectInput...");

	if(z_InitDirectInput() == false)
	{
		success = false;
		z_ReportProblem();
		if(z_win_err_debug)
			debug.Output("GameX: DirectInput init failed.\n");
	}
	else
	{
		if(z_win_log_debug)
			debug.Output("GameX: DirectInput init succeeded.\n");
	}

	if(z_win_request_no_sound)
	{
		z_mp3_support = false;
		z_mp3_support_tested = true;
	}

	z_DisplayLoadingWithValue (-5, "Initializing VSync...");

	z_InitVSync();

	z_DisplayLoadingWithValue (-5, "Testing Renderer...");

	return success;
}

void GameXClass::z_InitWindowsInstance (HINSTANCE hThisInst, int nWinMode)
{
	z_win_hinst = hThisInst;
	z_win_hmode = nWinMode;
}

extern LRESULT CALLBACK WindowFunc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

bool GameXClass::z_InitWindowClass ()
{
	WNDCLASS wcl;
	wcl.hCursor 		= LoadCursor(NULL, IDC_ARROW);
	wcl.hIcon			= LoadIcon(NULL, z_win_appIconIDStr);
	wcl.lpszMenuName	= "Menu" ;
	wcl.lpszClassName	= "GameX" ;
	wcl.hInstance		= z_win_hinst;
	wcl.style			= CS_VREDRAW | CS_HREDRAW;
	if(!z_win_request_no_mouse_input) wcl.style |= CS_DBLCLKS; // enable double-click recognition in the game window
	wcl.lpfnWndProc 	= (WNDPROC) WindowFunc;
	wcl.cbClsExtra		= 0;
	wcl.cbWndExtra		= 0;
	wcl.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
	z_win_resize_workaround = 0;

	if(!RegisterClass(&wcl)) { 
		z_win_status = GAMEX_INITCLASS_FAILED;
		return false;
	}
	z_win_status = GAMEX_INITCLASS_OK;
	return true;
}
/*
static unsigned long WINAPI PriorityManager ( void * ptd )
{
	HANDLE process = GetCurrentProcess();
	for(;;)
	{
		Sleep(2000); // this thread stays completely idle except for a little bit every 2 seconds:

		extern long z_lastActiveTime;
		if(clock() - z_lastActiveTime > 100) // if it's been more than 0.1 seconds since an active update,
		{
			if(GetPriorityClass(process) != NORMAL_PRIORITY_CLASS)
			{
				// maybe we froze - downgrade our priority or we'll take the whole system down with us
				// (either that or the window is just inactive or we're running too slowly, in which case we also won't mind downgrading)
				SetPriorityClass(process, NORMAL_PRIORITY_CLASS);
			}
		}
		else // if it's been less than 0.1 seconds since an active update,
		{
			if(GetPriorityClass(process) != HIGH_PRIORITY_CLASS)
			{
				// we're active and running acceptably well
				// give ourselves a high priority so music players and things won't destroy our nice framerate
				// NOTE: REALTIME_PRIORITY_CLASS works even better, except then the input firing can't catch up sometimes
				SetPriorityClass(process, HIGH_PRIORITY_CLASS);
			}
		}
	}
}
*/
bool GameXClass::z_InitWindows ()
{
	DWORD style;

	// Initialize View Format (Bits-Per-Pixel)
	if(z_win_request_bpp != 0)
		z_win_screenbpp = z_win_request_bpp; // set bpp to requested setting
	else
	{
		HDC hdc = GetDC (z_win_hwnd); // current setting requested
		z_win_screenbpp = GetDeviceCaps (hdc, BITSPIXEL);
		ReleaseDC (z_win_hwnd, hdc);

		// make it a supported mode (16-bit or 32-bit)
		if(z_win_request_bpp < 16) z_win_request_bpp = 16;
		if(z_win_request_bpp > 16) z_win_request_bpp = 32;
	}

	// Initialize View Size & FullScreen Mode
	z_win_screenx = GetSystemMetrics(SM_CXSCREEN);
	z_win_screeny = GetSystemMetrics(SM_CYSCREEN);
	z_win_fullscreen = z_win_request_full ? true : false;
	z_win_fake_fullscreen = z_win_request_soft_fullscreen ? true : (z_win_request_hard_fullscreen ? false : debug.IsRunningDebug());
	z_win_no_acc_err = false;

	if(z_win_request_xsize == 0 && z_win_request_ysize == 0)
	{
		z_win_request_xsize = z_win_screenx;
		z_win_request_ysize = z_win_screeny;

		if(z_win_request_xsize / 3 > z_win_request_ysize * 4)
			z_win_request_xsize = z_win_request_ysize * 4 / 3;
		else
			z_win_request_ysize = z_win_request_xsize * 3 / 4;
	}
	else
	{
		if(z_win_request_xsize == 0) z_win_request_xsize = z_win_screenx;
		if(z_win_request_ysize == 0) z_win_request_ysize = z_win_screeny;
	}

	bool invalid_size = false;
	if(z_win_request_xsize > 2048) z_win_request_xsize = 2048, invalid_size = true;
	if(z_win_request_ysize >= 2048) z_win_request_ysize = 2047, invalid_size = true;
	if(z_win_request_xsize < 64) z_win_request_xsize = 64, invalid_size = true;
	if(z_win_request_ysize < 64) z_win_request_ysize = 64, invalid_size = true;

	if(z_win_err_debug)
	{
		if(invalid_size)
		{
			char * str = "Warning: Invalid game window size requested.\nIt must be between 64x64 and 2048x2047 pixels.";
			debug.Output("%s\n",str);
			debug.Flush();
			MessageBox (z_win_hwnd, str, "GameX Notice", MB_ICONWARNING);
		}
	}

	// Initialize View Width & Height
	if(z_win_fullscreen == TRUE)	{
		style = WS_POPUP;	
		z_win_width = z_win_request_xsize;
		z_win_height = z_win_request_ysize;
	}
	else
	{
		style = WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX | WS_CLIPCHILDREN;
		if(z_win_request_can_resize) style |= WS_SIZEBOX | WS_MAXIMIZEBOX;
		z_win_width = z_win_request_xsize;		
		z_win_height = z_win_request_ysize;
	}

	z_win_realx = z_win_screenx;
	z_win_realy = z_win_screeny;
	int window_x = min(z_win_width, z_win_fullscreen ? z_win_screenx : z_win_realx-4); // temporary variable for actual width of window
	int window_y = min(z_win_height, z_win_fullscreen ? z_win_screeny : z_win_realy-32);

	// Initialize Drawing Viewport
	SetView(0,0,z_win_width,z_win_height);
		
	// Create window adjusting for desired client area size
	RECT	rect;
	int 	sx, sy;
	HWND	temp_hwnd;

	char title_str [256];
	sprintf(title_str, "%s - " GAMEX_VERSION_STRING, GetName());
///	sprintf(title_str, "%s", GetName());

	// Create temporary window to determine extended window size
	temp_hwnd = CreateWindowEx(WS_EX_APPWINDOW, "GameX", title_str, style, 0, 0, window_x, window_y, 0, 0, z_win_hinst, 0);
	if(temp_hwnd == NULL)
	{
		z_win_status = GAMEX_INITWINTEMP_FAILED;
		return false;		
	}
	z_win_status = GAMEX_INITWINTEMP_OK;
	GetClientRect (temp_hwnd, &rect);

	// set the rect to be the proper size and center it onscreen while we're at it
	sx = (window_x*2) - rect.right;
	sy = (window_y*2) - rect.bottom;
	if(z_win_fullscreen == TRUE)
	{
		rect.left = 0; rect.top = 0;
	}
	else
	{
		rect.left = (z_win_screenx/2) - (sx/2);
		rect.top = (z_win_screeny/2) - (sy/2);
	}	
	rect.right = rect.left + sx;		
	rect.bottom = rect.top + sy;

	// Create actual window (with correct size) which will be used

	z_win_hwnd = CreateWindowEx(
					WS_EX_APPWINDOW, 
					"GameX", 
					title_str, 
					style, 
					rect.left, rect.top, 
					rect.right-rect.left, rect.bottom-rect.top,
					0, 
					0, 
					z_win_hinst, 
					0);
	if(z_win_hwnd == NULL)
	{
		z_win_status = GAMEX_INITWINREAL_FAILED;		
		return false;
	}

/*
	// set up some process priority stuff
	freezeCheckThread = CreateThread(NULL, 512, PriorityManager, NULL, CREATE_SUSPENDED, NULL);
    if(freezeCheckThread)
	{
		ResumeThread(freezeCheckThread);
		SetThreadPriority(freezeCheckThread, NORMAL_PRIORITY_CLASS);
	}
*/
	SetProcessPriorityBoost(GetCurrentProcess(), TRUE);

	// Double check to make sure we have correct client rect size	
	GetClientRect (z_win_hwnd, &rect);	

	z_win_status = GAMEX_INITWINREAL_OK;
	return true;
}

bool GameXClass::z_WindowDraw ()
{
	ShowWindow (z_win_hwnd, z_win_hmode);
	if(UpdateWindow (z_win_hwnd) == false)
	{
		z_win_status = GAMEX_SHOWWIN_FAILED;		
		return false;
	}
	FlashWindow(z_win_hwnd,0);
	BringWindowToTop(z_win_hwnd);
	SetForegroundWindow(z_win_hwnd);
	z_win_status = GAMEX_SHOWWIN_OK;

	z_DisplayLoadingWithValue(5, "Initializing DirectDraw...");

	return true;
}

HRESULT CALLBACK DisplayEnumCallback (LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext)
{
	return GameX.z_HandleDisplayEnum(lpDDModeDesc, lpContext);
}

HRESULT GameXClass::z_HandleDisplayEnum (LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext)
{
	if(z_SupportsDepth(lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount)
	&& lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount == GetBPP()) { // if the bit depth matches
		if(lpDDModeDesc->dwWidth == GetWidth() && lpDDModeDesc->dwHeight == GetHeight()) { // if the resolution matches
			z_win_realx = GetWidth();
			z_win_realy = GetHeight();
			return DDENUMRET_CANCEL; // exact match, end enumeration with requested screen width and height
		}

		if((int)(lpDDModeDesc->dwWidth) >= GetWidth() && (int)(lpDDModeDesc->dwHeight) >= GetHeight() // if(this bigger than requested
		&& (int)(lpDDModeDesc->dwWidth) <= z_win_realx && (int)(lpDDModeDesc->dwHeight) <= z_win_realy // and this smaller than so far)
		|| (z_win_realx 				<  GetWidth() || z_win_realy					<  GetHeight()) // or (so far smaller than requested
		&& (int)(lpDDModeDesc->dwWidth) >= z_win_realx && (int)(lpDDModeDesc->dwHeight) >= z_win_realy) { // and this smaller than so far)
			z_win_realx = (int)(lpDDModeDesc->dwWidth);
			z_win_realy = (int)(lpDDModeDesc->dwHeight);
			return DDENUMRET_OK; // best match yet but not exact, remember this resolution and continue enumeration
		}
	}

	return DDENUMRET_OK; // rejected display mode, continue enumeration (or end automatically if there's no more to enumerate)
}

HRESULT CALLBACK DevicesEnumCallback (LPSTR lpDeviceDescription, LPSTR lpDeviceName, LPD3DDEVICEDESC7 lpD3DDeviceDesc, LPVOID lpContext)
{
	return GameX.z_HandleDeviceEnum(lpDeviceDescription, lpDeviceName, lpD3DDeviceDesc, lpContext);
}

HRESULT GameXClass::z_HandleDeviceEnum (LPSTR lpDeviceDescription, LPSTR lpDeviceName, LPD3DDEVICEDESC7 lpD3DDeviceDesc, LPVOID lpContext)
{
	if(IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DNullDevice))
		return DDENUMRET_OK;

	if(!GameX.z_SimulatingSoftwareOnly() && IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DTnLHalDevice) )
	{
		z_d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_CANCEL; // done
	}
	else if(!GameX.z_SimulatingSoftwareOnly() && IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DHALDevice) )
	{
		z_d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_OK;
	}
	else if(IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DRGBDevice))
	{
		z_d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_OK;
	}
	return D3DENUMRET_OK;
}

HRESULT CALLBACK ZBufferFormatEnumCallback (LPDDPIXELFORMAT lpDDPixFmt, LPVOID lpContext)
{
	return GameX.z_HandleZBufferFormatEnum(lpDDPixFmt, lpContext);
}

HRESULT GameXClass::z_HandleZBufferFormatEnum (LPDDPIXELFORMAT lpDDPixFmt, LPVOID lpContext)
{
	memcpy(z_d3d_zbuffer_format, lpDDPixFmt, sizeof(DDPIXELFORMAT));
	if(lpDDPixFmt != NULL && lpDDPixFmt->dwZBufferBitDepth == 16) return D3DENUMRET_CANCEL;

	return D3DENUMRET_OK;
}

bool GameXClass::z_InitDirectDraw ()
{
	z_dd_last_error = DD_OK;
	z_dd_was_lost = false;

	HRESULT status = DirectDrawCreateEx(NULL, (void **) &z_dd_main, IID_IDirectDraw7, NULL);

	if(FAILED(status))
	{
		z_DirectDrawError(status);
		z_win_status = GAMEX_DDSTART_FAILED;
		return false;
	}

	z_DisplayLoadingWithValue(-5, "Querying DirectDraw...");

	z_win_status = GAMEX_DDSTART_OK;

	// set drawing state values to default
	z_draw_state->src_rect.left = 0;
	z_draw_state->src_rect.top = 0;
	z_draw_state->src_rect.right = DEFAULT;
	z_draw_state->src_rect.bottom = DEFAULT;
	z_draw_state->colors = ColorX();
	z_draw_state->color_effect = D3DTOP_MODULATE;
	z_draw_state->angle = 0.0f;
	z_draw_state->scalex = 1.0f;
	z_draw_state->scaley = 1.0f;
	z_draw_state->transx = 0.0f;
	z_draw_state->transy = 0.0f;
	z_draw_state->depth2D = 1.0f;
	z_draw_state->warp_mode = 0;
	z_draw_state->flags = 0;
	FOR(i,4)
		z_draw_state->warp[i] = 1.0f;
	z_draw_render_batch_open = false;
	z_draw_3D_depth_sort = 0;
	z_memory_used_current = 0;
	z_memory_used_max = 0;

	DDCAPS caps;
	ZeroMemory(&caps,sizeof(caps));
	caps.dwSize = sizeof(caps);
	if(SUCCEEDED(z_dd_main->GetCaps(&caps, NULL)))
	{

		if(z_win_log_debug)
		{
			char stringy [64];
			sprintf(stringy,"GameX: Video memory available: %d megabytes.",caps.dwVidMemFree/1024/1024);
			debug.Output("%s\n",stringy);
		}

		int memfree = caps.dwVidMemFree;
		if(memfree < 1500000 && !z_win_request_software_only)
		{
			// if there are less than about 1.5 megabytes of VRAM available, assume there won't be much if any video acceleration
			char * str = "Warning: Lack of video acceleration detected!";
			MessageBox (z_win_hwnd, str, "GameX Notice", MB_ICONWARNING);
			debug.Output("%s\n",str);
	/// 	if(memfree == 0) z_win_request_software_only = true; // assume no hardware if there's no hardware memory at all
		}
		else
		{
			int memtotal = caps.dwVidMemTotal;
			if(memtotal < 22000000 && !z_win_request_software_only)
			{
				// if there are less than about 22 megabytes of VRAM total, assume the video card is probably too old (yes, the memory number only gives a very rough estimate of how good the card is)
				char * str = "Warning: Your video card appears to be outdated.\nConsider updating it if this game plays slowly or looks odd.";
				debug.Output("%s\n",str);
				debug.Flush();
				MessageBox (z_win_hwnd, str, "GameX Notice", MB_ICONWARNING);
			}
		}
	}
 
	if(z_InitDirectDrawMonitor() == false)
		return false;

	if(z_InitDirectDrawSurfaces() == false)
		return false;

	return true;
}

bool GameXClass::ToggleFullscreen(bool immediate)
{
	// Request full screen mode that's the opposite of the current one
	z_win_request_full = !z_win_fullscreen;

	if(z_win_log_debug)
		debug.Output("GameX: Toggling full-screen mode...\n");

	if(immediate)
	{
		z_win_fullscreen = !z_win_fullscreen; // change now in case of reentry during assert
		bool ok = z_ReInitScreen();

///		SetWarningOnDraw(false); // leave off for the cycle
///		z_GameDraw();

		return ok;
	}
	else
		return true;
}

bool GameXClass::ChangeGameSize(int xres, int yres)
{
	if(xres == z_win_width && yres == z_win_height)
		return true; // return if already set to size requested

	z_win_request_xsize = xres;
	z_win_request_ysize = yres;

	return z_ReInitScreen();
}

bool GameXClass::z_ReInitScreen()
{
	// Set the status to something else than GAMEX_READY so UpdateView knows to temporarily do nothing
	z_win_status = GAMEX_DDSTART_OK;
//	z_win_init_done = false;
	z_win_warning_on_draw = false;
	const bool vbufferWasEnabled = z_d3d_vbufferenabled;
	z_d3d_vbufferenabled = false;

	// Uninitialize the DirectX components that depend on the current window:

	// DirectInput
	SAFE_UNACQUIRE_RELEASE(z_di_mouse);
	SAFE_UNACQUIRE_RELEASE(z_di_keyboard);
	SAFE_RELEASE(z_di_main);

	// Direct3D
	SAFE_RELEASE(z_d3d_vbuffer);
	SAFE_RELEASE(z_d3d_device);	
	SAFE_RELEASE(z_d3d_main);

	// DirectDraw
	SAFE_RELEASE(z_dd_back_buf);
	SAFE_RELEASE(z_dd_front_buf);	
	// don't uninitialize z_dd_main or any sound/music stuff yet
	
	// Hide the window
	ShowWindow(z_win_hwnd,false);

	// Re-Initialize the window
	if(z_InitWindows() == false) return false;
	if(z_WindowDraw() == false) return false;

	// Re-Initialize the necessary DirectX components:
	if(z_InitDirectDrawMonitor() == false) return false;
	if(z_InitDirectDrawSurfaces() == false) return false;
	if(z_InitDirect3D() == false) return false;
	if(z_InitDirectInput() == false) return false;
	z_InitVSync();

	z_d3d_vbufferenabled = vbufferWasEnabled;

	SetForegroundWindow(z_win_hwnd);

	// Reset the GameX status to be ready and return true since we succeeded:
	z_win_status = GAMEX_READY;
	z_win_init_done = true;
	return true;
}

bool GameXClass::z_InitDirectDrawMonitor ()
{
	z_DisplayLoadingWithValue(-3, "Setting Monitor Level...");

	int style;
	if(z_win_fullscreen && !z_win_fake_fullscreen)
		style = DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT;
	else
		style = DDSCL_NORMAL;

	HRESULT status = z_dd_main->SetCooperativeLevel (z_win_hwnd, style);

	z_DisplayLoadingWithValue(-1, "Setting Monitor Level...");

	if(FAILED(status))
	{
		if(z_win_fullscreen && !z_win_fake_fullscreen)
		{
			style ^= (DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT); // try it without a couple flags
			status = z_dd_main->SetCooperativeLevel (z_win_hwnd, style);
			if(FAILED(status))
			{
				z_win_fake_fullscreen = true;
				///z_win_fullscreen = false; // try it in windowed mode (maybe this computer doesn't allow fullscreen)
				style = DDSCL_NORMAL;
				status = z_dd_main->SetCooperativeLevel (z_win_hwnd, style);
			}
		}
		if(FAILED(status)) { // if we still failed
			z_win_status = GAMEX_DDSETLEVEL_FAILED;
			return false;
		}
	}
	z_win_status = GAMEX_DDSETLEVEL_OK;

	z_DisplayLoadingWithValue(-1, "Setting Display Mode...");
/*
	// force-revert the screen depth if we're switching from soft fullscreen mode to windowed
	static int prev_screenx = GetSystemMetrics(SM_CXSCREEN); // must be static
	static int prev_screeny = GetSystemMetrics(SM_CYSCREEN);
	static int prev_screenbpp = -1;
	if(!z_win_fullscreen && z_win_fake_fullscreen && (prev_screenx != z_win_screenx || prev_screeny != z_win_screeny))
		status = z_dd_main->SetDisplayMode (prev_screenx, prev_screeny, prev_screenbpp,0,0);
	prev_screenx = GetSystemMetrics(SM_CXSCREEN);
	prev_screeny = GetSystemMetrics(SM_CYSCREEN);
	prev_screenbpp = z_win_screenbpp;

	// now set the display mode:
	if(z_win_fullscreen == true)
*/
	// make soft fullscreen mode even softer
	if(z_win_fullscreen == true && !z_win_fake_fullscreen)
	{
		z_win_realx = z_win_realy = 0;

		// enumerate available display modes and decide on one of them that best matches z_win_width, z_win_height, and z_win_screenbpp
		status = z_dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);

		z_DirectDrawError(status); // in case of error or to prevent compiler mishandling aliasing after callbacks

		if(!z_win_realx || !z_win_realy) { // if DisplayEnumCallback didn't set a screen size,
			
			// maybe the requested bit depth isn't supported, try for something higher
			while(z_win_screenbpp <= 32 && (!z_win_realx || !z_win_realy || !z_SupportsDepth(z_win_screenbpp)))
			{
				z_win_screenbpp += 8;
				status = z_dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				z_DirectDrawError(status);
			}
			// if that didn't work, try for something lower
			while(z_win_screenbpp > 0 && (!z_win_realx || !z_win_realy || !z_SupportsDepth(z_win_screenbpp)))
			{
				z_win_screenbpp -= 4;
				status = z_dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				z_DirectDrawError(status);
			}
			if(!z_win_realx || !z_win_realy)
			{
				z_win_status = GAMEX_SETMODE_FAILED;
				return false;
			}
		}
		
		status = z_dd_main->SetDisplayMode (z_win_realx, z_win_realy, z_win_screenbpp,0,0);

		z_win_screenx = GetSystemMetrics(SM_CXSCREEN);
		z_win_screeny = GetSystemMetrics(SM_CYSCREEN);

		if(FAILED(status))
		{
			z_DirectDrawError(status);
			z_win_status = GAMEX_SETMODE_FAILED;
			return false;
		}
		z_win_status = GAMEX_SETMODE_OK;

	}
	else
	{
		if(z_win_screenbpp != GetDeviceCaps (GetDC (z_win_hwnd), BITSPIXEL)) { // if we aren't in the desired bit mode
			if(z_SupportsDepth(z_win_screenbpp))
				status = z_dd_main->SetDisplayMode (z_win_screenx, z_win_screeny, z_win_screenbpp,0,0); // change to that many bpp
			else
				status = DDERR_GENERIC;
			if(FAILED(status))
			{
				// maybe the requested bit depth isn't supported, try for something higher
				while(z_win_screenbpp <= 32 && (FAILED(status) || !z_SupportsDepth(z_win_screenbpp)))
				{
					z_win_screenbpp += 8;
					status = z_dd_main->SetDisplayMode (z_win_screenx, z_win_screeny, z_win_screenbpp,0,0);
				}
				// if that didn't work, try for something lower
				while(z_win_screenbpp > 0 && (FAILED(status) || !z_SupportsDepth(z_win_screenbpp)))
				{
					z_win_screenbpp -= 4;
					status = z_dd_main->SetDisplayMode (z_win_screenx, z_win_screeny, z_win_screenbpp,0,0);
				}
				if(FAILED(status))
				{
					// try something hard-coded that should definitely work, as a last resort
					z_win_screenx = 640;
					z_win_screeny = 480;
					z_win_screenbpp = 16;
					status = z_dd_main->SetDisplayMode (z_win_screenx, z_win_screeny, z_win_screenbpp,0,0);
					if(FAILED(status))
					{
						z_win_status = GAMEX_SETMODE_FAILED;
						return false; // still failed... nothing else we can do in this case
					}
				}
			}
			UpdateWindow (z_win_hwnd); // redraw the window borders after switching or they'll all be black for a while
		}
		z_win_realx = min(z_win_width, z_win_realx-4); 
		z_win_realy = min(z_win_height, z_win_realy-32);

		z_win_status = GAMEX_SETMODE_OK;
	}

	if(z_win_fullscreen && z_win_fake_fullscreen)
	{
		MoveWindow(z_win_hwnd, 0,0, z_win_screenx, z_win_screeny, TRUE);
	}

	DWORD monitor_frequency = 0;
	z_dd_main->GetMonitorFrequency(&monitor_frequency);
	if(monitor_frequency == 0) monitor_frequency = 60; // just in case
	z_win_screenrefresh = monitor_frequency;
	z_win_vsync_enabled = false; // will enable later if we can

	if(z_win_log_debug)
		debug.Output("GameX: Chosen display mode: %dx%d %d-bit %dHz %s %sexclusive.\n", z_win_screenx, z_win_screeny, z_win_screenbpp, z_win_screenrefresh, z_win_fullscreen?"fullscreen":"windowed", z_win_fullscreen&&!debug.IsRunningDebug()?"":"non");

	z_win_xpos = 0;
	z_win_ypos = 0;

	return true;
}

bool GameXClass::z_InitDirectDrawSurfaces ()
{
	z_DisplayLoadingWithValue(-5, "Initializing Surface...");

	HRESULT status;
	DDSURFACEDESC2 ddsd;	
	POINT pt;

	SAFE_RELEASE(z_dd_back_buf);
	SAFE_RELEASE(z_dd_front_buf);	

	ZeroMemory (&ddsd, sizeof (ddsd));
	ddsd.dwSize = sizeof (ddsd);
 
	// Set front surface description
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	// Try to create primary surface 
	status = z_dd_main->CreateSurface (&ddsd, &z_dd_front_buf, NULL);
	if(FAILED(status)) {			
		z_DirectDrawError(status);
		z_win_status = GAMEX_FRONTSURFACE_FAILED; // No good, return error
		return false;
	}
	
	z_win_status = GAMEX_FRONTSURFACE_OK;
	// Set back surface description
	ddsd.dwSize = sizeof (ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
	if(z_win_request_software_only) ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth = z_win_width;
	ddsd.dwHeight = z_win_height;

	// Try to create secondary surface 
	status = z_dd_main->CreateSurface (&ddsd, &z_dd_back_buf, NULL);
	if(FAILED(status))
	{
		z_DirectDrawError(status);
		z_win_status = GAMEX_BACKSURFACE_FAILED; // No good, return error
		return false;
	}

	z_win_status = GAMEX_BACKSURFACE_OK;

	// Get Surface Description (pixel format, pitch, etc.)	
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_PITCH | DDSD_PIXELFORMAT;
	z_dd_front_buf->GetSurfaceDesc(&ddsd);
	if(ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY)
	{
		if(z_win_log_debug)
			debug.Output("GameX: Front buffer is in video memory.\n");
	}
	else
	{
		if(z_win_err_debug)
			debug.Output("GameX: Front buffer is NOT IN VIDEO MEMORY.\n");
	}
	z_dd_back_buf->GetSurfaceDesc(&ddsd);
	if(ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY)
	{
		z_win_back_buf_in_vid_mem = true;
		if(z_win_log_debug)
			debug.Output("GameX: Back buffer is in video memory.\n");
	}
	else
	{
		z_win_back_buf_in_vid_mem = false;
		if(z_win_err_debug)
			debug.Output("GameX: Back buffer is NOT IN VIDEO MEMORY.\n");
	}

	// Initialize Drawing Functions, Pixel Masks, Pitch and Width 
	// based on Surface Description 

	z_win_pitch = ddsd.lPitch;
	z_win_adjustedpitch = z_win_pitch * 8 / z_win_screenbpp;

	z_win_maskr = ddsd.ddpfPixelFormat.dwRBitMask;
	z_win_maskg = ddsd.ddpfPixelFormat.dwGBitMask;		
	z_win_maskb = ddsd.ddpfPixelFormat.dwBBitMask;
	
	z_win_shiftr = z_GetShift (z_win_maskr); // 11 for 16 bit, 10 for 15 bit, 16 for 24 or 32 bit
	z_win_shiftg = z_GetShift (z_win_maskg); // 5 for 16 or 15 bit, 8 for 24 or 32 bit
	z_win_shiftb = z_GetShift (z_win_maskb); // 0

	z_win_maxr = z_GetMax (z_win_maskr);
	z_win_maxg = z_GetMax (z_win_maskg);
	z_win_maxb = z_GetMax (z_win_maskb);

	z_win_maskminr = 256/(z_win_maskr>>z_win_shiftr);
	z_win_maskming = 256/(z_win_maskg>>z_win_shiftg);
	z_win_maskminb = 256/(z_win_maskb>>z_win_shiftb);

	z_NullDrawPixelFuncs();

	// now set a clipper to the back surface so we don't have to clip bitmap blits ourselves
	{
		RECT boundary[1] = {0,0,z_win_width,z_win_height};
		LPDIRECTDRAWCLIPPER clipper;
		LPRGNDATA rgn_data;
		if(FAILED(z_dd_main->CreateClipper(0,&clipper,NULL)))
		{
			z_win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		int rectSize = sizeof(RECT);
		int rgnHeaderSize = sizeof(RGNDATAHEADER);
		rgn_data = (LPRGNDATA)malloc(rgnHeaderSize+rectSize);
		memcpy(rgn_data->Buffer, boundary, rectSize);
		rgn_data->rdh.dwSize = rgnHeaderSize;
		rgn_data->rdh.nRgnSize = rectSize;
		rgn_data->rdh.nCount = 1;
		rgn_data->rdh.iType = RDH_RECTANGLES;
		rgn_data->rdh.rcBound = boundary[0];
		if(FAILED(clipper->SetClipList(rgn_data,0)) || FAILED(z_dd_back_buf->SetClipper(clipper)))
		{
			free(rgn_data);
			z_win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		free(rgn_data);
	}

	LPDIRECTDRAWCLIPPER clipper = NULL; // hold the clipper
	z_dd_main->CreateClipper(0,&clipper,NULL);
	clipper->SetHWnd(0, z_win_hwnd);
	z_dd_front_buf->SetClipper(clipper);

	// Precompute: y-table look up 
	for(int y = 0; y <= z_win_height; y++)
		z_win_ytable[y] = y*z_win_adjustedpitch;
		
	// Precompute: screen size (total number of bytes)
	z_win_size = z_win_pitch*z_win_height;
	
	// Precompute: client rectangle
	
	RECT destrect;

	GetClientRect(z_win_hwnd, &destrect);

	z_win_realx = destrect.right - destrect.left; // necessary on occasions where the OS gives us a different window size
	z_win_realy = destrect.bottom - destrect.top; // than we requested, such as when the requested width is very small or very large

	if(!z_IsResizeWorkaroundEnabled())
	{
		z_SetProportional();
	}
	else
	{
		z_win_xoffset = (z_win_realx - z_win_width) / 2;
		z_win_yoffset = (z_win_realy - z_win_height) / 2; 			
		z_win_realx = z_win_width;
		z_win_realy = z_win_height;
	}

	pt.x = z_win_xpos;
	pt.y = z_win_ypos;
	ClientToScreen(z_win_hwnd, &pt);
	OffsetRect(&destrect, pt.x, pt.y);
	z_win_xpos = pt.x;
	z_win_ypos = pt.y;
	
	z_win_active = true;	
	z_win_status = GAMEX_READY;

	// keep the cursor inside the game if we're in fullscreen or were requested to keep it in the window
	if(z_win_fullscreen || z_win_request_keep_cursor_in_window)
	{
		RECT rect = {z_win_xpos+z_win_xoffset, z_win_ypos+z_win_yoffset, z_win_xpos+z_win_xoffset+z_win_realx, z_win_ypos+z_win_yoffset+z_win_realy} ;
		ClipCursor(&rect);
	}
	else
	{
		ClipCursor(NULL);
	}

	ClearScreen();
	UpdateView();

	return true;
}

void GameXClass::z_InitVSync ()
{
	if(z_win_request_vsync && z_win_fps > z_win_screenrefresh || z_win_fps == -1)
		z_win_fps = z_win_screenrefresh; // set the frame rate to not exceed the refresh rate

	z_win_vsync_enabled = z_win_request_vsync ? true : false;

	if(z_win_log_debug)
	{
		debug.Output("GameX: Refresh synchronization %s.\n", z_win_request_vsync ? "on" : "OFF");
		debug.Output("GameX: Timer resolution: %I64d ticks per second.\n", GetClocksPerSec());
	}
}

bool GameXClass::z_InitDirect3D ()
{
	z_d3d_vbufferenabled = false;

	// decide on a max texture size we'll allow for D3D surfaces
	// 1024 is a good limit; any higher wastes much speed and memory
	// note that it may also be further limited by hardware capability
	z_SetImageResolutionLimit (1024);

	// if DirectDraw is working, use it to intiialize Direct3D:
	if(z_dd_main == NULL || FAILED(z_dd_main->QueryInterface(IID_IDirect3D7, (LPVOID *)& z_d3d_main)))
	{
		if(z_win_err_debug)
			debug.Output("Direct3D: Device query failed.\n");
		z_win_status = GAMEX_D3DSTART_FAILED;
		return false; 
	}


	HRESULT status;

	// Enumerate available 3D rendering devices and decide on one:
	status = z_d3d_main->EnumDevices((LPD3DENUMDEVICESCALLBACK7) &DevicesEnumCallback, NULL);
	GUID guid = z_d3d_device_desc->deviceGUID;

	z_d3d_zbuffer_enabled = false;
	z_d3d_zbuffer_wasEnabled = false;

	// make sure not to use a z-buffer unless we have access to video hardware
	if(guid != IID_IDirect3DRGBDevice)
	{

		DDPIXELFORMAT pixelzbuffer;
		z_d3d_main->EnumZBufferFormats(guid, (LPD3DENUMPIXELFORMATSCALLBACK) &ZBufferFormatEnumCallback, (VOID*)&pixelzbuffer );

		DDSURFACEDESC2 ddsdZ;
		ZeroMemory(&ddsdZ, sizeof(DDSURFACEDESC2));
		ddsdZ.dwSize = sizeof(DDSURFACEDESC2);
		ddsdZ.dwWidth = z_win_width;
		ddsdZ.dwHeight = z_win_height;
		ddsdZ.ddpfPixelFormat = *z_d3d_zbuffer_format;
		ddsdZ.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_CAPS | DDSD_PIXELFORMAT;
		ddsdZ.ddsCaps.dwCaps = DDSCAPS_ZBUFFER;
		ddsdZ.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY; // require the z-buffer to be in video memory, otherwise it would be too slow

/// 	// used when software z-buffering is allowed (in which case the previous line should be disabled instead):
/// 	if(!z_win_back_buf_in_vid_mem || (z_d3d_device_desc->deviceGUID == IID_IDirect3DRGBDevice || z_d3d_device_desc->deviceGUID == IID_IDirect3DRefDevice))
/// 		ddsdZ.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		status = z_dd_main->CreateSurface(&ddsdZ, &z_d3d_zbuffer, NULL);
		if(SUCCEEDED(status))
		{
			status = z_dd_back_buf->AddAttachedSurface(z_d3d_zbuffer);
			if(FAILED(status))
			{
				if(z_win_err_debug)
					debug.Output("GameX: Failed to attach z buffer.\n");
			} else
				z_d3d_zbuffer_enabled = true;
		}
		else
		{
			if(z_win_err_debug)
				debug.Output("Warning: Z-Buffering not supported on this computer.\n");
		}
	}
	// if z buffering hasn't been enabled at this point, depth sorting is used instead
	// (note that depth sorting is used even with z-buffering when drawing translucently)


	// make sure we assign a 3D device to the back buffer first,
	// or it probably won't get 3D hardware acceleration
	if(z_CreateDeviceForSurface(z_d3d_device, z_dd_back_buf) == false)
	{
		z_win_status = GAMEX_D3DSTART_FAILED;
		return false;
	}


	// Set up the description of the vertex buffer
	D3DVERTEXBUFFERDESC vBufferDesc;
	ZeroMemory(&vBufferDesc, sizeof(D3DVERTEXBUFFERDESC));
	vBufferDesc.dwSize = sizeof(D3DVERTEXBUFFERDESC);
	vBufferDesc.dwNumVertices = D3DMAXNUMVERTICES;
//	vBufferDesc.dwFVF =  D3DFVF_TLVERTEX;
	vBufferDesc.dwFVF = FVF_TLVERTEX;
	vBufferDesc.dwCaps |= D3DVBCAPS_WRITEONLY;
	vBufferDesc.dwCaps |= D3DVBCAPS_DONOTCLIP;
//	vBufferDesc.dwCaps |= D3DVBCAPS_OPTIMIZED;
	if(guid == IID_IDirect3DRGBDevice)
		vBufferDesc.dwCaps |= D3DVBCAPS_SYSTEMMEMORY;

	// make the vertex buffer
	HRESULT vbhr = z_d3d_main->CreateVertexBuffer(&vBufferDesc,&z_d3d_vbuffer, D3DDP_WAIT | D3DDP_DONOTCLIP | D3DDP_DONOTUPDATEEXTENTS | D3DDP_DONOTLIGHT);
	z_d3d_vbuffermaxverts = vBufferDesc.dwNumVertices;

	z_LockVBuffer();

	// set some render and texture state values, most of these may be the defaults, but just in case:

	z_d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP); // we don't want texture wrapping,
	z_d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_CLAMP); // it doesn't work right on most image sizes anyway
	z_d3d_device->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_MODULATE);
	z_d3d_device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);
	z_d3d_device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_LINEAR);

	z_d3d_device->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE); // we have no D3D light sources, for now
	z_d3d_device->SetRenderState(D3DRENDERSTATE_SPECULARENABLE, FALSE); // don't need specular
	z_d3d_device->SetRenderState(D3DRENDERSTATE_COLORVERTEX, TRUE); // we want to allow per-vertex coloring
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); // we want to allow alpha blending
	z_d3d_device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE); // and dithering, except when temporarily disabled later
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ANTIALIAS, FALSE); // set antialiasing off
	z_d3d_device->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW);
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,FALSE); // the z buffer is only enabled while drawing in 3D
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE); 
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHAREF, 248);
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATER);

	SetView(z_win_viewx1,z_win_viewy1,z_win_viewx2,z_win_viewy2); // now that D3D has been initialized set the view again for it

	z_draw_render_batch_device = NULL;

	if(z_d3d_zbuffer_enabled)
		z_d3d_device->Clear(0,NULL,D3DCLEAR_ZBUFFER,0,1.0f,0);

	return true;
}

bool GameXClass::z_CreateDeviceForSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface)
{
	if(z_win_request_software_only || FAILED(z_d3d_main->CreateDevice(IID_IDirect3DTnLHalDevice, surface, &device))) { // the best 3D hardware device
		if(z_win_request_software_only || FAILED(z_d3d_main->CreateDevice(IID_IDirect3DHALDevice, surface, &device))) { // second-best 3D hardware
			if(FAILED(z_d3d_main->CreateDevice(IID_IDirect3DMMXDevice, surface, &device))) { // best 3D software device
				// failing up until now is normal, although undesirable
				HRESULT status = z_d3d_main->CreateDevice(IID_IDirect3DRGBDevice, surface, &device); // backup 3D software
				if(FAILED(status))
				{
					// that last try should never fail, something's very wrong if we get here
					z_DirectDrawError(status);
					z_Direct3DError(status);
					z_win_status = GAMEX_D3DSTART_FAILED;
					if(z_win_err_debug)
						debug.Output("Direct3D: Device create failed.\n");
					return false; // should never get here; even if there's no 3D device, the D3DRGB device is software emulated
				}
			}
		}
	}
	return true;
}

void GameXClass::z_DisplayLoadingWithValue (int percent, const char * message)
{
	if(percent < 0)
		percent = z_lastPercent - percent;
	z_lastPercent = percent;
	if(percent > 100)
		percent = 0;

	char str [64];
	sprintf(str, "Loading %d%%...", percent);

	ClearScreen();
	if(percent)
		DrawGDIText(z_win_viewx2-120,z_win_viewy2-30,str);
	if(message)
		DrawGDIText(18,z_win_viewy2-30,message);

	UpdateView();

	if(!z_win_hwnd || z_win_hwnd == GetForegroundWindow())
	{
		if(GetAsyncKeyState(VK_ESCAPE) & 0x8000
		|| ((GetAsyncKeyState(VK_MENU) & 0x8000) && (GetAsyncKeyState(VK_F4) & 0x8000)))
			TerminateProcess(GetCurrentProcess(), 0); // extreme prejudice
	}

	MSG msg;
	while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
	{
		if(msg.message == WM_QUIT) TerminateProcess(GetCurrentProcess(), 0); // extreme prejudice;
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

// gets amount to shift a color component of the given mask
int GameXClass::z_GetShift (int mask)
{
	if(mask == 0) return 0;
	int count;
	for(count = 0; (mask & 1) == 0; count++)
		mask >>= 1;
	return count;
}

int GameXClass::z_GetMax (int mask)
{
	if(mask == 0) return 0;
	int count, max; 
	for(count = 0; (mask & 1) == 0; count++)
		mask >>= 1;
	for(max = 0; mask > 0; max++)
		mask >>= 1;
	return max;
}


bool GameXClass::z_InitDirectInput ()
{
	// Create the DirectInput object:

	if(FAILED(DirectInputCreate(z_win_hinst, DIRECTINPUT_VERSION, &z_di_main, NULL)))
		return false;

	bool failed = false;

	// Create and set up the keyboard device:

	z_di_keyboard = NULL;
	if(SUCCEEDED(z_di_main->CreateDevice(GUID_SysKeyboard, &z_di_keyboard, NULL)))
	{
		// set up support for buffering
		{
			DIPROPDWORD  dipdw; 
			dipdw.diph.dwSize = sizeof(DIPROPDWORD); 
			dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
			dipdw.diph.dwObj = 0; 
			dipdw.diph.dwHow = DIPH_DEVICE; 
			dipdw.dwData = DX_KEY_BUFSIZE; 
			if(FAILED(z_di_keyboard->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph)))
				if(z_win_err_debug)
					debug.Output("GameX: Enabling DirectInput keyboard buffering failed.\n");
		}

		HRESULT status = DDERR_GENERIC;
		if(!z_win_request_no_auto_pause)
		{
			status = z_di_keyboard->SetCooperativeLevel(z_win_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND | DISCL_NOWINKEY);
			if(SUCCEEDED(status))
				status = z_di_keyboard->SetDataFormat(&c_dfDIKeyboard);
			if(SUCCEEDED(status))
				status = z_di_keyboard->Acquire();
		}
		if(FAILED(status))
		{
			status = z_di_keyboard->SetCooperativeLevel(z_win_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
			if(SUCCEEDED(status))
				status = z_di_keyboard->SetDataFormat(&c_dfDIKeyboard);
			if(SUCCEEDED(status))
				status = z_di_keyboard->Acquire();
		}
		if(FAILED(status))
		{
			status = z_di_keyboard->SetCooperativeLevel(z_win_hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
			if(SUCCEEDED(status))
				status = z_di_keyboard->SetDataFormat(&c_dfDIKeyboard);
			if(SUCCEEDED(status))
				status = z_di_keyboard->Acquire();
		}
		if(FAILED(status))
			failed = true;
		else
		{
			FOR(k,255)
				z_di_keystate[k] = z_di_keypressed[k] = 0;
		}

	} else
		failed = true;

	z_di_mouse = NULL;
	if(!z_win_request_no_mouse_input)
	{

		// Create and set up the mouse device:

		if(SUCCEEDED(z_di_main->CreateDevice(GUID_SysMouse, &z_di_mouse, NULL)))
		{
			// set up support for buffering
			{
				DIPROPDWORD  dipdw; 
				dipdw.diph.dwSize = sizeof(DIPROPDWORD); 
				dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
				dipdw.diph.dwObj = 0; 
				dipdw.diph.dwHow = DIPH_DEVICE; 
				dipdw.dwData = DX_KEY_BUFSIZE; 
				if(FAILED(z_di_mouse->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph)))
					if(z_win_err_debug)
						debug.Output("GameX: Enabling DirectInput mouse buffering failed.\n");
			}

			HRESULT status = DDERR_GENERIC;
			if(!debug.IsRunningDebug())
			{
				status = z_di_mouse->SetCooperativeLevel(z_win_hwnd, ((z_win_fullscreen && !z_win_request_no_hide_cursor) ? DISCL_EXCLUSIVE : DISCL_NONEXCLUSIVE) | DISCL_FOREGROUND); // "the exclusive setting causes the Microsoft Windows cursor to disappear"
				if(SUCCEEDED(status))
					status = z_di_mouse->SetDataFormat(&c_dfDIMouse);
				if(SUCCEEDED(status))
					status = z_di_mouse->Acquire();
				if(FAILED(status))
				{
					status = z_di_mouse->SetCooperativeLevel(z_win_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
					if(SUCCEEDED(status))
						status = z_di_mouse->SetDataFormat(&c_dfDIMouse);
					if(SUCCEEDED(status))
						status = z_di_mouse->Acquire();
				}
			}
			if(FAILED(status))
			{
				status = z_di_mouse->SetCooperativeLevel(z_win_hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND); // "DISCL_BACKGROUND enables you to debug the application more easily, but at a cost in performance"
				if(SUCCEEDED(status))
					status = z_di_mouse->SetDataFormat(&c_dfDIMouse);
				if(SUCCEEDED(status))
					status = z_di_mouse->Acquire();
			}

			if(FAILED(status))
				failed = true;
		}
		else
		{
			failed = true;
			z_win_request_no_mouse_input = true;
		}
	} 

	z_UpdateInput(); // so the game can check for initial input inside GameInit() right after calling GameX.Initialize()

	return !failed;
}



bool GameXClass::z_InitDirectSound ()
{
	if(z_sound_support_tested || z_win_request_no_sound) return true;

	assertMsg(z_win_hwnd, "Trying to use DirectSound before GameX has been initialized.");
	if(!z_win_hwnd) return false;

	z_sound_support_tested = true;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Initializing DirectSound...");

	if(FAILED(DirectSoundCreate (NULL, &z_ds_main, NULL)))
	{
		z_ds_main = NULL;
		z_win_status = GAMEX_SOUNDINIT_FAILED;
		z_sound_support = false;
	}
	else if(FAILED(z_ds_main->SetCooperativeLevel(z_win_hwnd, DSSCL_PRIORITY)))
	{
		SAFE_RELEASE(z_ds_main);
		z_win_status = GAMEX_SOUNDCOOP_FAILED;
		z_sound_support = false;
	}
	else
	{
		z_sound_support = true;

		if(!z_win_init_done)
			z_DisplayLoadingWithValue (-5, "Loading Game...");
		return true;
	}

	z_ReportProblem();
	if(z_win_err_debug)
		debug.Output("GameX: DirectSound init failed.\n");

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Loading Game...");
	return false;
}

bool GameXClass::z_InitDirectMusicPerformance ()
{
	if(z_win_request_no_sound || z_dm_performance) return true;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Initializing DirectMusic Performance...");

	z_InitDirectSound();

	bool ok = true;
	// set up DirectMusic and performance/ports:

	if(FAILED(CoCreateInstance(CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC,
								IID_IDirectMusicPerformance, (void**)&z_dm_performance)))
		ok = false;

	if(ok && FAILED(z_dm_performance->Init(NULL, (z_sound_support ? z_ds_main : NULL), z_win_hwnd)))
		ok = false;

	if(ok && FAILED(z_dm_performance->AddPort(NULL)))
		ok = false;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Loading Game...");

	return ok;
}

bool GameXClass::z_InitDirectMusicLoader ()
{
	if(z_win_request_no_sound || z_dm_loader) return true;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Initializing DirectMusic Loader...");

	// prepare DirectMusic for loading MIDI files:

	bool ok = true;
	if(FAILED(CoCreateInstance(CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC,
								IID_IDirectMusicLoader, (void**)&z_dm_loader)))
		ok = false;

	z_dm_musicx_index = 0;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Loading Game...");

	return true;
}

bool GameXClass::z_TestDirectShow ()
{
	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Testing DirectShow...");

	IGraphBuilder * dsh_gb;
	HRESULT status = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC, IID_IGraphBuilder, (void **)&dsh_gb);
	if(FAILED(status))
	{
		z_mp3_support = false;
///		z_ReportProblem();
		if(z_win_err_debug)
			debug.Output("GameX: DirectShow init failed.\n");
	}
	else
	{
		SAFE_RELEASE(dsh_gb);
		z_mp3_support = true;
		if(z_win_log_debug)
			debug.Output("GameX: DirectShow init succeeded.\n");
	}

	z_mp3_support_tested = true;

	if(!z_win_init_done)
		z_DisplayLoadingWithValue (-5, "Loading Game...");

	return z_mp3_support;
}



LPDIRECTDRAWSURFACE7 GameXClass::z_CreateImageSurface (int width, int height, bool texture, int create_alpha, bool targetable)
{
	if(z_dd_main == NULL)
	{
		z_ReportProblem(GAMEX_NO_INIT); // loading image before GameX was initialized
	}

	LPDIRECTDRAWSURFACE7 surface; // the surface we'll create and return

	DDSURFACEDESC2 ddsd;
	ZeroMemory (&ddsd, sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(ddsd);

	DDPIXELFORMAT ddpf;
	ZeroMemory (&ddpf, sizeof(DDPIXELFORMAT));
	ddpf.dwSize = sizeof(ddpf);

	ddsd.dwHeight = height;
	ddsd.dwWidth = width;
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0; // specify black as transparent if drawn masked
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if(texture && z_d3d_device != NULL)
	{
		ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
		D3DDEVICEDESC7 ddDesc;
		z_d3d_device->GetCaps(&ddDesc);
		if(ddDesc.deviceGUID == IID_IDirect3DHALDevice || ddDesc.deviceGUID == IID_IDirect3DTnLHalDevice)
		{
			ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
		}
		else
		{
			ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		}

		if(targetable) { // && texture
			ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE; // to allow the surface to request a 3D device
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTDYNAMIC; // could speed up images that are frequently drawn to
		} else
		{
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTSTATIC; // could speed up images that are never drawn to
		}

		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2)
		{
			// if the texture dimensions must be a power of 2,
			// increase the surface size accordingly (the change is transparent to the user and programmer)
			for(ddsd.dwWidth  = 1; (int)ddsd.dwWidth  < width ; ddsd.dwWidth  <<= 1);
			for(ddsd.dwHeight = 1; (int)ddsd.dwHeight < height; ddsd.dwHeight <<= 1);
		}
		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY || targetable) { // for some reason DDSCAPS_3DDEVICE causes non-square textures to fail drawing
			// if the texture has to be square
			if(ddsd.dwWidth > ddsd.dwHeight) ddsd.dwHeight = ddsd.dwWidth;
			else							 ddsd.dwWidth  = ddsd.dwHeight;
		}
		int maxWidth = min((DWORD)z_d3d_max_tex_width, ddDesc.dwMaxTextureWidth);
		int maxHeight = min((DWORD)z_d3d_max_tex_height, ddDesc.dwMaxTextureHeight);
		if(ddsd.dwWidth  > (DWORD)maxWidth)  ddsd.dwWidth  = (DWORD)maxWidth;
		if(ddsd.dwHeight > (DWORD)maxHeight) ddsd.dwHeight = (DWORD)maxHeight;
		if(z_win_request_low_resolution) ddsd.dwWidth /= 2, ddsd.dwHeight /= 2;
		if(ddsd.dwWidth  < ddDesc.dwMinTextureWidth)  ddsd.dwWidth  = ddDesc.dwMinTextureWidth;
		if(ddsd.dwHeight < ddDesc.dwMinTextureHeight) ddsd.dwHeight = ddDesc.dwMinTextureHeight;
	} else
	{
		// if we're not making a 3D accelerated surface, just try for a 2D accelerated one
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	}

			if(create_alpha)
			{
				z_dd_back_buf->GetPixelFormat(&ddpf);
				ddpf.dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
				ddpf.dwRGBBitCount	   = 32;
				ddpf.dwRGBAlphaBitMask = 0xFF000000;
				ddpf.dwRBitMask 	   = 0x00FF0000;
				ddpf.dwGBitMask 	   = 0x0000FF00;
				ddpf.dwBBitMask 	   = 0x000000FF;
				ddsd.dwFlags |= DDSD_PIXELFORMAT;
				ddsd.ddpfPixelFormat = ddpf;
			} 
			
	HRESULT status = z_dd_main->CreateSurface(&ddsd, &surface, NULL);
	if(FAILED(status))
	{
		if(ddsd.dwFlags & DDSD_PIXELFORMAT)
		{
			ddsd.dwFlags ^= DDSD_PIXELFORMAT; // try again, this time accepting any format it wants to use
			status = z_dd_main->CreateSurface(&ddsd, &surface, NULL);
		}
		if(FAILED(status))
		{
			z_DirectDrawError(status);
			if(z_win_err_debug)
				debug.Output("Failed to allocate image memory.\n");
			Quit(true);
			return NULL;
		}
	}

	if(z_win_status != GAMEX_NO_INIT)
	{
		surface->GetPixelFormat(&ddpf); // get the pixel format DirectDraw made the surface in,
		int bytes = (ddsd.dwWidth*ddsd.dwHeight*ddpf.dwRGBBitCount/8);
		int size = bytes/1024; if(size < 1) size = 1;
		if(z_win_log_debug)
		{
			if(z_GetStatus() == GAMEX_READY)
			{
				char sizeStr [8];
				sprintf(sizeStr, "%05d", size);
				for(int z = 0 ; sizeStr[z] == '0' ; z++) sizeStr[z] = ' ';

				debug.Output("%s KB, %dx%d%s -> %dx%dx%d\n",sizeStr,width,height,(create_alpha)?"+A":"", ddsd.dwWidth,ddsd.dwHeight,ddpf.dwRGBBitCount);
			}
		}
		z_MemoryRecordChange(bytes);
	}

	return surface;
}

bool GameXClass::z_DirectXCleanUp ()
{
	// DirectMusic MIDI songs:
	while(z_dm_musicx_index--)
		if(z_dm_musicx[z_dm_musicx_index] != NULL)
			z_dm_musicx[z_dm_musicx_index]->Release(),	z_dm_musicx[z_dm_musicx_index] = NULL;

	// DirectMusic
	if(z_dm_performance) z_dm_performance->CloseDown(), z_dm_performance->Release(), z_dm_performance = NULL;
	if(z_dm_loader)		z_dm_loader->Release(),		 z_dm_loader = NULL;
	z_mp3_support_tested = false;
	
	// DirectSound
	{
		unsigned int size = (unsigned int)(z_sound_list.size());
		for(unsigned int i = 0 ; i < size ; ++i)
			z_sound_list[i]->Unload();
		z_sound_list.clear();
	}
	SAFE_RELEASE(z_ds_main);
	z_sound_support_tested = false;

	// DirectInput
	SAFE_UNACQUIRE_RELEASE(z_di_mouse);
	SAFE_UNACQUIRE_RELEASE(z_di_keyboard);
	SAFE_RELEASE(z_di_main);

	// Direct3D
	SAFE_RELEASE(z_d3d_vbuffer);
	SAFE_RELEASE(z_d3d_device);	
	SAFE_RELEASE(z_d3d_main);

	// DirectDraw
	SAFE_RELEASE(z_dd_back_buf);
	SAFE_RELEASE(z_dd_front_buf);	
	SAFE_RELEASE(z_dd_main);

	CoUninitialize();
	return true;
}

void GameXClass::z_CleanUp ()
{
	if(z_win_log_debug)
	{
		debug.Output("Maximum Graphic/Sound memory in use: %d bytes\n", z_memory_used_max);
		debug.Output("Graphic/Sound memory in use on exit: %d bytes\n", z_memory_used_current);
	}
	if(z_DirectXCleanUp() == FALSE)
	{
		if(z_win_err_debug)
			debug.Output("GameX::CleanUp: Unable to properly shut down DirectX.\n");
	}
}

void GameXClass::z_LoadQueuedImages ()
{
	if(z_queued_image_loads.empty())
		return;

	bool wasDone = z_win_init_done;
	z_win_init_done = false;

	size_t size = z_queued_image_loads.size();
	int i = 0;
	int prevPercent = z_lastPercent;

//	int modCheck = max(1, (int)logf((float)size+1.0f));
	int modCheck = max(4, size/50);

	while(!z_queued_image_loads.empty())
	{
		ImageX* image = z_queued_image_loads.back();
		z_queued_image_loads.pop_back();
		if(image)
			image->ForceLoad();

		if((i % modCheck == 0 || i == size-1) && !z_win_init_done)
		{
			char str [32];
			sprintf(str, "Loading Images %d%%...", 100 * (i+1) / size);
			z_DisplayLoadingWithValue(prevPercent + (100 - prevPercent) * (i+1) / size, str);
		}
		i++;
	}
	z_queued_image_loads.clear();

	z_win_init_done = wasDone;
}
